# =========================================================================== #
from typing import Annotated, Self

import yaml
from pydantic import Field

# --------------------------------------------------------------------------- #
from captura.config import BaseHashable
from captura.config import Config as ConfigCaptura
from captura.schemas import mwargs


class DummyItemConfig(BaseHashable):
    minimum: Annotated[int, Field()]
    maximum: Annotated[int, Field()]


class DummyGrantConfig(BaseHashable):
    maximum_self: Annotated[int, Field(default=144)]
    minimum_self: Annotated[
        int,
        Field(
            default=81,
            gt=24,
            description=(
                "Minimum number of grants to create for a particular "
                "user. Must be greater than 24 so that all users have a "
                "chance of having documents in all possible permutations of"
                "``(pending, level, PendingFrom.granter | PendingFrom.grantee,"
                " deleted)``. When this occurs, an errors is not thrown but "
                "instead a warning is produced in logs."
            ),
        ),
    ]

    minimum_other: Annotated[int, Field(default=64)]
    maximum_other: Annotated[int, Field(default=81)]


class DummyUserConfig(BaseHashable):
    # minimum_id: Annotated[int, Field(default=50)]
    minimum: Annotated[
        int,
        Field(
            default=125,
            description=(
                "Minimum number of users that ``DummyHandler`` will maintain "
                "in the database."
            ),
        ),
    ]
    maximum_uses: Annotated[
        int | None,
        Field(
            default=100,
            description=(
                "Number of times that a dummy may be used before it is "
                "disposed. Note that ``tainted`` dummies will still be "
                "disposed of. Generally this number should be high if tests"
                "are to complete quickly. When ``None``, dummies have no use "
                "limit."
            ),
        ),
    ]


# NOTE: Inherits from :class:`BaseHashable` because this will be used to
#       overwrite ``Depends(config)``.
class DummyConfig(BaseHashable):

    # model_config = YamlSettingsConfigDict(yaml_files=util.Path.config("dummy.yaml"))
    users: Annotated[
        DummyUserConfig,
        Field(default_factory=lambda: mwargs(DummyUserConfig)),
    ]
    documents: Annotated[
        DummyItemConfig,
        Field(
            default_factory=lambda: mwargs(
                DummyItemConfig,
                maximum=36,
                minimum=25,
            )
        ),
    ]
    collections: Annotated[
        DummyItemConfig,
        Field(
            default_factory=lambda: mwargs(
                DummyItemConfig,
                maximum=9,
                minimum=16,
            )
        ),
    ]
    grants: Annotated[
        DummyGrantConfig,
        Field(default_factory=lambda: mwargs(DummyGrantConfig)),
    ]

    @classmethod
    def load(cls, manifest_path: str) -> Self:
        with open(manifest_path, "r") as file:
            raw = yaml.safe_load(file)

        return cls.model_validate(raw)


class ConfigSimulatus(ConfigCaptura):
    dummy: DummyConfig



"""This module contains all of the depends that will be needed in variou
endpoints.

The code behind most of the functions here should be in other modules, for
instance `auth` or `config`, except in the case that it is dependency used to
trap and transform api parameters. No exports (things in ``__all__``) should
be wrapped in ``Depends``.
"""

# =========================================================================== #
from functools import cache
from typing import Annotated, TypeAlias

from fastapi import Depends, Header, HTTPException, Request
from sqlalchemy import select
from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio.session import AsyncSession, async_sessionmaker
from sqlalchemy.orm import Session, sessionmaker

# --------------------------------------------------------------------------- #
from captura import util
from captura.auth import Auth, Token, TokenPermissionTier, try_decode
from captura.config import Config
from captura.controllers.access import Access
from captura.controllers.create import Create, Update
from captura.controllers.delete import Delete
from captura.controllers.read import Read
from captura.models import User
from captura.views.args import QueryForce

# NOTE: `cache` is used instead of using the `use_cache` keyword argument of
#       `Depends` because it will result in identical object ids. For instance
#       this means that `session_maker` will always return the same session
#       maker.

logger = util.get_logger(__name__)


@cache
def config() -> Config:
    """The configuration dependency function.

    This function should only run once. I do not know if it is better to use
    ``functolls.cache`` or to use the ``cache`` keyword of ``fastapi.Depends``.
    The logging statement is used to make sure that the internals of this
    function are only called once.

    :returns: Application configuration. See :class:`Config`.
    """
    logger.debug("Dependency `config` called.")
    return Config()  # type: ignore


_DependsConfig = Depends(config, use_cache=False)
DependsConfig: TypeAlias = Annotated[Config, _DependsConfig]


@cache
def engine(config: DependsConfig) -> Engine:
    """The engine dependency function.

    Note that bringing in a database connector is a supported dependency
    pattern, for instance look at this:

    .. code:: txt

        https://fastapi.tiangolo.com/tutorial/sql-databases/

    :param config: Configuration specifying the MySQL host.
    :returns: An ``sqlalchemy.engine.Engine`` instance connected to the
        host as specified by :param:`config`.
    """
    logger.debug("Dependency `engine` called.")
    return config.engine()


@cache
def async_engine(config: DependsConfig) -> AsyncEngine:
    logger.debug("Dependency `engine` called.")
    return config.async_engine()


DependsEngine: TypeAlias = Annotated[Engine, Depends(engine, use_cache=False)]
DependsAsyncEngine: TypeAlias = Annotated[
    AsyncEngine, Depends(async_engine, use_cache=False)
]


@cache
def session_maker(engine: DependsEngine) -> sessionmaker[Session]:
    """
    :param engine: An ``sqlalchemy.engine``, probably from :func:`engine`.
    :returns: A ``sqlalchemy.orm.sessionmaker`` for creating sessions.
    """
    logger.debug("Dependency `session_maker` called.")
    return sessionmaker(engine)


@cache
def async_session_maker(
    async_engine: DependsAsyncEngine,
) -> async_sessionmaker[AsyncSession]:
    """
    :param engine: An ``sqlalchemy.engine``, probably from :func:`engine`.
    :returns: A ``sqlalchemy.orm.sessionmaker`` for creating sessions.
    """
    logger.debug("Dependency `session_maker` called.")
    return async_sessionmaker(bind=async_engine, class_=AsyncSession)


DependsSessionMaker: TypeAlias = Annotated[
    sessionmaker[Session],
    Depends(session_maker),
]
DependsAsyncSessionMaker: TypeAlias = Annotated[
    async_sessionmaker[AsyncSession],
    Depends(async_session_maker),
]


# =========================================================================== #
# Authentication dependencies.


@cache
def auth(config: DependsConfig) -> Auth:
    """
    :param: Application configuration. Specifies ``auth0`` configuration.
    :returns: The authentication handler defined in :module:`.auth` for
        ``auth0`` specified by :param:``.
    """

    logger.debug("Dependency `auth` called.")
    return (Auth.forAuth0 if config.auth0.use else Auth.forPyTest)(config)


DependsAuth: TypeAlias = Annotated[Auth, Depends(auth, use_cache=False)]

HeaderAuthorization = Annotated[str, Header(description="Auth0 bearer token.")]
HeaderAuthorizationOptional = Annotated[
    str | None, Header(description="Optional auth0 bearer token.")
]


@cache
def token(
    auth: DependsAuth,
    authorization: HeaderAuthorization,
) -> Token:
    """Decode and deserialize the bearer JWT specified in the ``Authorization``
    header.

    Database validation happens later once a database session is available.

    :param config: Application configuration.
    :param auth: Auth0 or PyTest authorization handler.
    :param authorization: Authorization header. It should match
        :const:`PATTERN_BEARER`.
    """
    decoded, err = try_decode(auth, authorization)
    if err is not None:
        raise err
    token = Token.model_validate(decoded)
    return token


def token_optional(
    auth: DependsAuth,
    authorization: HeaderAuthorizationOptional = None,
) -> Token | None:
    if authorization is not None:
        return token(auth, authorization)
    return None


def token_admin(_token: "DependsToken"):
    token = Token.model_validate(_token)
    if token.tier != TokenPermissionTier.admin:
        raise HTTPException(403, detail="Admin only.")

    return token


DependsTokenOptional: TypeAlias = Annotated[
    None | Token,
    Depends(token_optional),
]
DependsToken: TypeAlias = Annotated[Token, Depends(token)]
DependsTokenAdmin: TypeAlias = Annotated[Token, Depends(token_admin)]


def uuid(token: DependsToken, uuid: str | None = None) -> str:
    """Get the user UUID either from the token or get the uuid from query
    parameters.

    Be careful when using this on endpoints that specific to one user.
    """
    return uuid if uuid is not None else token.subject


DependsUUID: TypeAlias = Annotated[str, Depends(uuid)]


# NOTE: Caching should make it such that this function is not invoked multiple
#       times when making multiple requests. If the same JWT is decoded,
#       verified, and deserialized on each request by a user a certain overhead
#       will be added to every request when this might not be necessary.
#
#       There is a potential problem with caching, which is that multiple user
#       tokens will be persisted in memory, which could be a vulnerability that
#       might allow 'intruders' into deployments to potentially get a hold of
#       user tokens. Of course, if such a 'hacker' had access to the
#       deployment, the game is over.
@cache
def user(
    config: DependsConfig,
    sessionmaker: DependsSessionMaker,
    token: DependsToken,
) -> User:
    """ """
    uuid = token.subject
    if uuid is None:
        raise HTTPException(
            401,
            detail="Invalid token: Token does not specify a user uuid.",
        )

    logger.debug("Dependency `auth` called for user with uuid `%s`.", uuid)

    with sessionmaker() as session:
        user = session.execute(select(User).where(User.uuid == uuid)).scalar()

        if user is None:
            raise HTTPException(
                401,
                detail="Invalid Token: User does not exist.",
            )

    return user


DependsUser: TypeAlias = Annotated[User, Depends(user)]

# --------------------------------------------------------------------------- #


def access(
    sessionmaker: DependsSessionMaker,
    token: DependsTokenOptional,
    request: Request,
):
    logger.debug("Creating sessionmaker access.")
    with sessionmaker() as session:
        yield Access(session=session, token=token, method=request.method)
        logger.debug("Exitting `sessionmaker` context.")


DependsAccess: TypeAlias = Annotated[Access, Depends(access)]


def create(
    token: DependsTokenOptional,
    access: DependsAccess,
    request: Request,
    force: QueryForce = False,
) -> Create:
    api_origin = request.url.path
    return Create(
        force=force,
        session=access.session,
        token=token,
        method=request.method,
        api_origin=api_origin,
    )


def read(
    token: DependsTokenOptional,
    access: DependsAccess,
    request: Request,
) -> Read:
    return Read(session=access.session, token=token, method=request.method)


def api_origin(request: Request) -> str:
    return f"{request.method} {request.url.path}"


DependsApiOrigin = Annotated[str, Depends(api_origin)]


def update(
    token: DependsTokenOptional,
    access: DependsAccess,
    request: Request,
    api_origin: DependsApiOrigin,
    force: QueryForce = False,
) -> Update:
    return Update(
        session=access.session,
        token=token,
        method=request.method,
        api_origin=api_origin,
        force=force,
    )


def delete(
    token: DependsTokenOptional,
    access: DependsAccess,
    request: Request,
    api_origin: DependsApiOrigin,
    force: QueryForce = False,
) -> Delete:
    return Delete(
        session=access.session,
        token=token,
        method=request.method,
        api_origin=api_origin,
        access=access,
        force=force,
    )


DependsCreate: TypeAlias = Annotated[Create, Depends(create)]
DependsRead: TypeAlias = Annotated[Read, Depends(read)]
DependsUpdate: TypeAlias = Annotated[Update, Depends(update)]
DependsDelete: TypeAlias = Annotated[Delete, Depends(delete)]



"""Note: None of these schemas should include ``id`` fields. Instead they
should return err.UUID fields. This is done because, by design, the err.UUIDS of
objects must known if they are to be got. Take for instance the following flow:

1. The ``GET /user`` endpoint gets a users err.UUID from their JWT. The uuid is
    sent in the response.
2. The ``GET /user/collections`` endpoint returns the collections for the user
   returned from the above request.
3. The ``GET /collections/<collection_id>/documents`` endpoint can be used to
   get some err.UUIDs to make queries to the document collection.

From a mathmatical standpoint, this is 'good enough' because err.UUIDs should map
uniquely to a corresponding database row (further, for tables with multiple
foreign keys it is not necessary to specify multiple values).
"""

# =========================================================================== #
import enum
from datetime import datetime, timedelta
from typing import (
    Annotated,
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    Self,
    Set,
    Type,
    TypeVar,
)

from fastapi import Query
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic_core.core_schema import FieldValidationInfo
from rich.align import Align, AlignMethod, VerticalAlignMethod
from rich.json import JSON
from rich.panel import Panel

# --------------------------------------------------------------------------- #
from captura import fields
from captura.util import check_enum_opt_attr


# NOTE: Could use partials but I like this pattern more.
def create_validate_datetime(delta: timedelta) -> BeforeValidator:
    def validator(v: Any, field_info: FieldValidationInfo) -> Any:
        if v:
            return v

        utcnow = datetime.now()
        return utcnow - delta

    return BeforeValidator(validator)


# --------------------------------------------------------------------------- #


class KindSchema(str, enum.Enum):
    search = "search"
    metadata = "metadata"
    default = "default"
    extra = "extra"
    create = "create"
    update = "update"


class KindNesting(str, enum.Enum):
    array = "list"
    object = "dict"


# --------------------------------------------------------------------------- #


class Registry:
    schemas: Dict[fields.FieldKindObject, Dict[KindSchema, Type["BaseSchema"]]]

    def __init__(self):
        self.schemas = dict()

    def add(self, schema: Type["BaseSchema"]):
        kind, kind_schema = schema.kind_mapped, schema.kind_schema
        schemas = self.schemas

        if kind not in schemas:
            schemas[kind] = {kind_schema: schema}
            return

        schemas_for_kind = schemas[kind]
        if kind_schema in schemas_for_kind:
            msg = f"Registry already has value for `{kind} -> {kind_schema}`."
            msg += f"`(existing value = {schemas_for_kind[kind_schema]}`"
            msg += f"`overwritting_value = {schema}`)."
            raise ValueError(msg)

        schemas_for_kind[kind_schema] = schema

    def get(
        self,
        kind: fields.KindObject,
        kind_schema: KindSchema,
    ) -> "Type[BaseSchema]":
        schemas = self.schemas
        if (schemas_for_kind := schemas.get(kind)) is None:
            raise ValueError(f"No schemas for kind `{kind.name}`.")
        elif (schema := schemas_for_kind.get(kind_schema)) is None:
            msg = f"No schema of type `{kind_schema}` for `{kind.name}`."
            raise ValueError(msg)
        return schema


registry = Registry()


class BaseSchema(BaseModel):
    model_config = ConfigDict(use_enum_values=False, from_attributes=True)

    # NOT FIELDS SINCE THEY ARE CONSTANT. METADATA FOR CONSUMERS!
    kind_mapped: ClassVar[fields.FieldKindObject]
    kind_schema: ClassVar[KindSchema]
    registry: ClassVar[Registry] = registry
    registry_exclude: ClassVar[bool] = False

    def __init_subclass__(cls) -> None:
        if "Base" in cls.__name__:
            return

        if not cls.registry_exclude:
            check_enum_opt_attr(
                cls,
                "kind_mapped",
                fields.KindObject,
            )
            check_enum_opt_attr(cls, "kind_schema", KindSchema)
            registry.add(cls)

    def __rich__(self) -> Panel:
        return self.render()

    def render(
        self,
        align: AlignMethod = "center",
        vertical: VerticalAlignMethod = "middle",
        **kwargs,
    ) -> Panel:
        jj = JSON(self.model_dump_json(**kwargs))
        return Panel(Align(jj, align, vertical=vertical))


class BaseSearchSchema(BaseSchema):
    kind_mapped: ClassVar[fields.FieldKindObject]
    kind_schema: ClassVar[KindSchema]

    # NOTE: It appears that using `Query(None)` as the field default is what
    #       will help fastapi find `uuid_event`. This was tricky to find so do
    #       not touch this. The result was though up on the basis that using
    #       `uuid_event: Set[str] | None = Query(None)` in the endpoint
    #       signature worked.
    uuids: Annotated[Set[str] | None, Field(default=Query(None))]
    limit: fields.FieldLimitOptional
    name_like: fields.FieldNameLike
    description_like: fields.FieldDescriptionLike
    include_public: Annotated[bool, Field(default=True)]
    randomize: Annotated[bool, Field(default=False)]


class BaseUpdateSchema(BaseSchema):
    """This exists because the editor feedback from the decorator feedback was
    not very good at all. Unfortunately all of the fields for update requests
    will have to be explicit.
    """

    kind_mapped: ClassVar[fields.FieldKindObject]
    kind_schema: ClassVar[KindSchema]

    @model_validator(mode="after")
    def at_least_one(self) -> Self:
        if not any(getattr(self, field) is not None for field in self.model_fields):
            msg = ", ".join((str(vv) for vv in self.model_fields.keys()))
            raise ValueError(f"Must specify at least one of `{msg}`.")
        return self


class BasePrimarySchema(BaseSchema):
    public: bool = True


class BaseSecondarySchema(BaseSchema): ...


class BasePrimaryTableExtraSchema(BaseSchema):
    id: fields.FieldID
    deleted: fields.FieldDeleted


# --------------------------------------------------------------------------- #


class UserBaseSchema(BasePrimarySchema):
    # _mapped_class = User
    kind_mapped = fields.KindObject.user

    email: fields.FieldEmail
    name: fields.FieldName
    description: fields.FieldDescription
    url_image: fields.FieldUrl
    url: fields.FieldUrl


class UserCreateSchema(UserBaseSchema):
    kind_schema = KindSchema.create

    content: fields.FieldContent


class UserUpdateSchema(BaseUpdateSchema):
    kind_mapped = fields.KindObject.user
    kind_schema = KindSchema.update

    # NOTE: `url_image` and `url` already optional.
    name: Optional[fields.FieldName] = None
    description: Optional[fields.FieldDescription] = None
    url_image: fields.FieldUrl = None
    url: fields.FieldUrl = None

    content: fields.FieldContent


class UserSchema(UserBaseSchema):
    kind_schema = KindSchema.default

    uuid: fields.FieldUUID

    content: fields.FieldContent


class UserExtraSchema(BasePrimaryTableExtraSchema, UserSchema):
    kind_schema = KindSchema.extra

    deleted: fields.FieldDeleted


class UserSearchSchema(BaseSearchSchema):
    kind_mapped = fields.KindObject.user
    kind_schema = KindSchema.search


# =========================================================================== #
# Collection and Assignment Schema


class CollectionBaseSchema(BasePrimarySchema):
    kind_mapped = fields.KindObject.collection

    name: fields.FieldName
    description: fields.FieldDescription


class CollectionCreateSchema(CollectionBaseSchema):
    kind_schema = KindSchema.create

    content: fields.FieldContent


class CollectionUpdateSchema(BaseUpdateSchema):
    kind_mapped = fields.KindObject.collection
    kind_schema = KindSchema.update

    uuid_user: Optional[fields.FieldUUID] = None
    name: Optional[fields.FieldName] = None
    description: Optional[fields.FieldDescription] = None

    content: fields.FieldContent


class CollectionMetadataSchema(CollectionBaseSchema):
    kind_schema = KindSchema.metadata
    uuid_user: fields.FieldUUID
    uuid: fields.FieldUUID


class CollectionSchema(CollectionMetadataSchema):
    kind_schema = KindSchema.default

    content: fields.FieldContent


class CollectionExtraSchema(BasePrimaryTableExtraSchema, CollectionSchema):
    kind_schema = KindSchema.extra


class CollectionSearchSchema(BaseSearchSchema):
    kind_schema = KindSchema.search
    kind_mapped = fields.KindObject.collection


# =========================================================================== #
# Assignments


class AssignmentBaseSchema(BaseSecondarySchema):
    kind_mapped = fields.KindObject.assignment


# NOTE: NO UPDATE SCHEMA! UPDATING IS NOT ALLOWED. MOST FIELDS NOT UPDATABLE
class AssignmentCreateSchema(AssignmentBaseSchema):
    kind_schema = KindSchema.create


class AssignmentSchema(AssignmentBaseSchema):
    kind_schema = KindSchema.default

    uuid: fields.FieldUUID
    uuid_collection: fields.FieldUUID
    uuid_document: fields.FieldUUID


class AssignmentExtraSchema(AssignmentSchema):
    kind_schema = KindSchema.extra

    deleted: fields.FieldDeleted
    id_collection: fields.FieldID
    id_document: fields.FieldID


# =========================================================================== #
# Documents


class DocumentBaseSchema(BasePrimarySchema):
    kind_mapped = fields.KindObject.document

    name: fields.FieldName
    description: fields.FieldDescription


class DocumentCreateSchema(DocumentBaseSchema):
    kind_schema = KindSchema.create

    content: fields.FieldContent = None


class DocumentUpdateSchema(BaseUpdateSchema):
    kind_mapped = fields.KindObject.document
    kind_schema = KindSchema.update

    name: Optional[fields.FieldName] = None
    description: Optional[fields.FieldDescription] = None

    content: fields.FieldContent = None


class DocumentMetadataSchema(DocumentBaseSchema):
    kind_schema = KindSchema.metadata

    uuid: fields.FieldUUID


class DocumentSchema(DocumentMetadataSchema):
    kind_schema = KindSchema.default

    content: fields.FieldContent


class DocumentExtraSchema(BasePrimaryTableExtraSchema, DocumentSchema):
    kind_schema = KindSchema.extra


class DocumentSearchSchema(BaseSearchSchema):
    kind_mapped = fields.KindObject.document
    kind_schema = KindSchema.search


class TimespanLimitParams(BaseModel):
    limit: fields.FieldLimitOptional
    before: fields.FieldUnixTimestampOptional
    after: fields.FieldUnixTimestampOptional

    # noqa: prop-decorator
    @property
    def before_timestamp(self) -> int | None:
        if self.before is None:
            return None
        return int(datetime.timestamp(self.before))

    # noqa: prop-decorator
    @property
    def after_timestamp(self) -> int | None:
        if self.after is None:
            return None
        return int(datetime.timestamp(self.after))

    @model_validator(mode="after")
    def check_before_after(self) -> Self:
        if self.before is None or self.after is None:
            return self

        if self.before < self.after:
            raise ValueError("`before` must be less than `after`.")

        return self


# =========================================================================== #
# Grants


class GrantBaseSchema(BaseSecondarySchema):
    kind_mapped = fields.KindObject.grant

    # NOTE: `uuid_document` is not included here because this is only used in
    #       `POST /grants/users/<uuid>`.
    level: fields.FieldLevel

    @field_serializer("level")
    def enum_as_name(item: enum.Enum):  # type: ignore
        return item.name


# NOTE: NO UPDATE SCHEMA! UPDATING IS NOT ALLOWED. MOST FIELDS NOT UPDATABLE
class GrantCreateSchema(GrantBaseSchema):
    kind_schema = KindSchema.create


class GrantSchema(GrantBaseSchema):
    kind_schema = KindSchema.default

    # Useful
    uuid: fields.FieldUUID
    uuid_document: fields.FieldUUID
    uuid_user: fields.FieldUUID
    pending: fields.FieldPending
    pending_from: fields.FieldPendingFrom

    @field_serializer("level", "pending_from")
    def enum_as_name(item: enum.Enum):  # type: ignore
        return item.name

    # Metadata
    uuid_parent: Optional[fields.FieldUUID] = None
    uuid_user_granter: Optional[fields.FieldUUID] = (
        None  # should it reeally be optional
    )


class GrantExtraSchema(GrantSchema):
    kind_schema = KindSchema.extra

    deleted: fields.FieldDeleted
    id_document: fields.FieldID
    id_user: fields.FieldID
    pending: fields.FieldPending
    pending_from: fields.FieldPendingFrom

    children: Annotated["List[GrantExtraSchema]", Field()]


# =========================================================================== #
# Events Schema


class EventBaseSchema(BaseSchema):
    kind_mapped = fields.KindObject.event

    api_origin: str
    api_version: str
    uuid_parent: fields.FieldUUIDOptional
    uuid: fields.FieldUUID
    uuid_obj: fields.FieldUUID
    uuid_user: fields.FieldUUID
    kind: fields.FieldKindEvent
    kind_obj: fields.FieldKindObject
    timestamp: fields.FieldUnixTimestamp
    detail: fields.FieldDetail

    # noqa: prop-decorator
    @property
    def timestamp_string(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp)  # type: ignore

    @field_serializer("kind_obj", return_type=str)
    def serailize__obj(self, v: Any, info) -> str:
        return v.name

    @field_validator("kind_obj", mode="before")
    def validate_kind(
        cls, v: None | str | fields.FieldKindObject
    ) -> None | fields.FieldKindObject:
        if isinstance(v, str):
            try:
                w = fields.KindObject[v]
            except KeyError:
                w = fields.KindObject._value2member_map_.get(v)  # type: ignore
                if w is None:
                    msg = f"Could not find enum value associated with `{v}`."
                    raise ValueError(msg)
            return w

        return v


class EventMetadataSchema(EventBaseSchema):
    kind_schema = KindSchema.metadata


class EventSchema(EventMetadataSchema):
    kind_schema = KindSchema.default

    children: Annotated["List[EventSchema]", Field(default=list())]


class EventExtraSchema(EventBaseSchema):
    kind_schema = KindSchema.extra

    data: "AsOutput"
    children: Annotated["List[EventExtraSchema]", Field(default=list())]


class BaseEventParams(BaseModel):
    limit: fields.FieldLimitOptional
    before: fields.FieldUnixTimestampOptional
    after: Annotated[
        datetime,
        Field(validate_default=True, default=None),
        create_validate_datetime(timedelta(days=3)),
    ]

    def _timestamp(self, v: datetime | None) -> int | None:
        return int(datetime.timestamp(v)) if v is not None else None

    @property
    def timestamp_before(self) -> int | None:
        return self._timestamp(self.before)

    @property
    def timestamp_after(self) -> int | None:
        return self._timestamp(self.after)


class EventSearchSchema(BaseEventParams):
    kind_mapped: ClassVar[fields.FieldKindObject] = fields.KindObject.event
    kind_schema: ClassVar[KindSchema] = KindSchema.search

    # NOTE: It appears that using `Query(None)` as the field default is what
    #       will help fastapi find `uuid_event`. This was tricky to find so do
    #       not touch this. The result was though up on the basis that using
    #       `uuid_event: Set[str] | None = Query(None)` in the endpoint
    #       signature worked.
    uuid_event: Annotated[Set[str] | None, Field(default=Query(None))]

    kind: fields.FieldKindEvent
    kind_obj: fields.FieldKindObject
    uuid_obj: fields.FieldUUID | None = None


class EventParams(BaseEventParams):
    # NOTE: https://docs.pydantic.dev/2.0/usage/types/datetime/
    root: bool = True


# =========================================================================== #
# Composite schemas
#
# NOTE: Some of these might fit into the previous categories.


T_Output = TypeVar(
    "T_Output",
    # User
    UserSchema,
    UserExtraSchema,
    List[UserSchema],
    List[UserExtraSchema],
    # Document
    DocumentSchema,
    DocumentMetadataSchema,
    DocumentExtraSchema,
    List[DocumentMetadataSchema],
    List[DocumentSchema],
    List[DocumentExtraSchema],
    # Collection
    CollectionSchema,
    CollectionMetadataSchema,
    CollectionExtraSchema,
    List[CollectionSchema],
    List[CollectionMetadataSchema],
    List[CollectionExtraSchema],
    # Grant
    GrantSchema,
    GrantExtraSchema,
    List[GrantSchema],
    List[GrantExtraSchema],
    # Assignment
    AssignmentSchema,
    AssignmentExtraSchema,
    List[AssignmentSchema],
    List[AssignmentExtraSchema],
    # Event
    EventMetadataSchema,
    EventSchema,
    EventExtraSchema,
    List[EventMetadataSchema],
    List[EventSchema],
    List[EventExtraSchema],
)


# This is the primary response. See https://youtu.be/HBH6qnj0trU?si=7YIqUkPl4gB5S_sP
class AsOutput(BaseModel, Generic[T_Output]):
    # : Annotated[fields.FieldKindObject, BeforeValidator(kind), Field()]
    data: Annotated[T_Output, Field(desciption="Wrapped data.")]

    # noqa: prop-decorator
    @property
    def kind(self) -> fields.FieldKindObject | None:
        if (ff := self.first()) is not None:
            return ff.kind_mapped
        return None

    # noqa: prop-decorator
    @property
    def kind_schema(self) -> KindSchema | None:
        if (ff := self.first()) is not None:
            return ff.kind_schema
        return None

    # noqa: prop-decorator
    @property
    def kind_nesting(self) -> KindNesting | None:
        match self.data:
            case list():
                return KindNesting.array
            case dict():
                return KindNesting.object
            case _:
                return None

    # noqa: prop-decorator
    @property
    def name_schema(self) -> str:
        return self.first().__class__.__name__

    def first(self) -> Any | None:
        match self.data:
            case [object() as item, *_]:
                return item
            case []:
                return None
            case object() as item:
                return item
            case bad:
                msg = f"Cannot determine `` from malformed data `{bad}`."
                raise ValueError(msg)


# T_OutputEvents = TypeVar(
#     "T_OutputEvents",
#     EventSchema,
#     EventWithObjectsSchema,
# )


class OutputWithEvents(AsOutput, Generic[T_Output]):
    events: Annotated[List[EventSchema], Field()]


# --------------------------------------------------------------------------- #


class TMwargs(Protocol):
    # NOTE: Adding a signature for init with ``kwargs`` is a pain.
    ...


T_mwargs = TypeVar("T_mwargs", bound=BaseModel | TMwargs)


# NOTE: Cause I hate wrapping kwargs in dict and pydantic constructor hints are
#       often annoying. This can be applied anything with a kwargs constructor.
def mwargs(M: Type[T_mwargs], **kwargs) -> T_mwargs:
    return M(**kwargs)  # type: ignore



"""Api routers and functions.
This includes a metaclass so that undecorated functions may be tested.

"""

# =========================================================================== #
import enum
from http import HTTPMethod
from os import path
from typing import Any, ClassVar, Dict, Literal

from fastapi import APIRouter
from fastapi.templating import Jinja2Templates

# --------------------------------------------------------------------------- #
from captura import util
from captura.err import AnyErrDetailAccessDocumentGrant, ErrDetail, ErrObjMinSchema

logger = util.get_logger(__name__)
# logger.level = logging.INFO


class OpenApiTags(str, enum.Enum):
    users = "users"
    documents = "documents"
    collections = "collections"
    edits = "edits"
    grants = "grants"
    assignments = "assignments"
    events = "events"

    # etc
    auth0 = "auth0"
    admin = "admin"
    html = "html"


OpenApiTagMetadata = [
    {
        "name": "users",
        "description": "View and Manage Users.",
    },
    {
        "name": "documents",
        "description": "View and Manage Documents.",
    },
    {
        "name": "collections",
        "description": "View and Manage Collections.",
    },
    {
        "name": "grants",
        "description": "View and manage document access and access invitations.",
    },
    {
        "name": "assignments",
        "description": "View and manage assignments of documents to collections.",
    },
    {
        "name": "events",
        "description": "View and manage object events.",
    },
    {"name": "auth0", "description": "Auth0 and token tools."},
    {
        "name": "html",
        "description": "For humans.",
    },
]

OpenApiResponseCommon = {
    404: dict(
        model=ErrDetail[ErrObjMinSchema],
        description="Object could not be found.",
    ),
    410: dict(
        model=ErrDetail[ErrObjMinSchema],
        detail="Object is deleted or is pending deletion.",
    ),
}
OpenApiResponseUnauthorized = {
    401: dict(
        model=ErrDetail[Literal["Token required"]],
    ),
}
OpenApiResponseDocumentForbidden = {
    403: dict(
        model=AnyErrDetailAccessDocumentGrant,
        description=(
            "For read, cannot access document because no grants exist and "
            "the document is private. Otherwise grants do not exist or they "
            "are pending."
        ),
    ),
}


class ViewMixins:
    """

    :attr view_children: Dictionary of instances to instances.
    :attr view_router: The router built by :class:`ViewMeta`.
    :attr view: Mapping from router function names to router routes.
    """

    # view_children: ClassVar[Dict[str, Type]] = dict()
    view_children: ClassVar[Dict[str, "ViewMeta"]] = dict()
    view_router: ClassVar[APIRouter]
    view_router_args: ClassVar[Dict[str, Any]] = dict()  # type: ignore
    view_routes: ClassVar[Dict[str, str | Dict[str, Any]]] = dict()  # type: ignore
    view_templates: ClassVar[Jinja2Templates] = Jinja2Templates(
        directory=path.join(path.dirname(__file__), "templates")
    )


class ViewMeta(type):
    """Metaclass to handle routing.

    It will build a router under `view`.
    """

    @classmethod
    def add_route(cls, T, fn_name: str, fn_info_raw: str | Dict[str, Any]):
        name = T.__name__

        # NOTE: Annotation is stange bc of the following mypy error:
        #       Incompatible types in capture pattern (pattern captures type "dict[object, object]", variable has type "dict[str, Any]")  [misc]
        info: Dict[Any, Any]
        url: str
        match fn_info_raw:
            case str() as url:
                info = dict()
            case {"url": url, **info}:
                ...
            case bad:
                msg = "Invalid info for url: Expected `str` for url or "
                msg += f"`dict` specifying atleast a url, got `{bad}`."
                raise ValueError(msg)

        # Parse name
        raw, _ = fn_name.split("_", 1)
        http_meth = next((hh for hh in HTTPMethod if hh.value.lower() == raw), None)
        if http_meth is None:
            logger.warning(f"Could not determine method of `{fn_name}`.")
            return

        # Update status code if not provided.
        if http_meth == HTTPMethod.POST and "status" not in info:
            info.update(status_code=201)

        # Find attr
        fn = getattr(T, fn_name, None)
        if fn is None:
            msg = f"No such method `{fn_name}` of `{name}`."
            raise ValueError(msg)

        # Get the decoerator and call it.
        logger.debug("Adding function `%s` at url `%s`.", fn.__name__, url)
        decorator = getattr(T.view_router, http_meth.value.lower())
        decorator(url, **info)(fn)

    def __new__(cls, name, bases, namespace):
        T = super().__new__(cls, name, bases, namespace)
        logger.debug("Validating `%s` router.", name)

        # Validate `view_children`.
        if not hasattr(T, "view_children"):
            raise ValueError(f"`{name}` must define `view_children`.")
        elif not isinstance(T.view_children, dict):  # type: ignore
            raise ValueError(f"`{name}.view_children` must be a `dict`.")

        # Validate `view`.
        if not hasattr(T, "view_routes"):
            raise ValueError(f"`{name}` must define `view`.")
        elif not isinstance(T.view_routes, dict):  # type: ignore
            raise ValueError(f"`{name}.view` must be a dict.")

        # Validate `view_router_args`.
        if not hasattr(T, "view_router_args"):
            raise ValueError(f"`{name}` must define `view_router_args`.")
        elif not isinstance(T.view_router_args, dict):  # type: ignore
            raise ValueError(f"`{name}.view_router_args` must be a `dict`.")

        if name != "BaseView":
            # Create router.
            logger.debug("Creating router for `%s`.", name)
            T.view_router = (  # type: ignore
                T.view_router  # type: ignore
                if hasattr(T, "view_router")
                else APIRouter(**T.view_router_args)  # type: ignore
            )
            for fn_name, fn_info in T.view_routes.items():  # type: ignore
                cls.add_route(T, fn_name, fn_info)

            for child_prefix, child in T.view_children.items():  # type: ignore
                logger.debug(
                    "Adding child router `%s` for `%s`.",
                    child_prefix,
                    name,
                )
                T.view_router.include_router(  # type: ignore
                    child.view_router,
                    prefix=child_prefix,
                )

        return T


class BaseView(ViewMixins, metaclass=ViewMeta): ...



# =========================================================================== #
from os import environ
from typing import Annotated, Any, Dict, List, Optional

import typer
import yaml
from pydantic import SecretStr

# --------------------------------------------------------------------------- #
from captura import util
from captura.schemas import mwargs
from legere import flags, hooks
from legere.config import Config, HostConfig, ProfileConfig
from legere.handlers import CONSOLE, ConsoleHandler, HandlerData
from legere.requests import Requests
from legere.requests.assignments import AssignmentRequests
from legere.requests.base import BaseRequests, BaseTyperizable, ContextData, params
from legere.requests.collections import CollectionRequests
from legere.requests.documents import DocumentRequests
from legere.requests.grants import GrantRequests
from legere.requests.tokens import TokenRequests
from legere.requests.users import UserRequests

if util.PATH_HOOKS_USE:
    hooks.do_hooks(Requests)


FlagConfigOut = Annotated[
    Optional[str],
    typer.Option(
        "--config-out",
        "-o",
        help=(
            "Create/overwrite specified configuration file with output that "
            "would go the terminal (minus secret seconsoring)."
        ),
    ),
]
FlagConfig = Annotated[
    Optional[str],
    typer.Option(
        "--config",
        help=(
            "Configuration file to add to and load from. Overwrites the "
            "global `--config` option. To specify output, see `--config-out` "
            "if it is provided."
        ),
    ),
]
ArgumentProfileName = Annotated[
    str, typer.Argument(help="Name of the profile to update.")
]
FlagProfileCurrent = Annotated[
    bool, typer.Option("--current/--all", help="Show the current profile.")
]
FlagProfileNames = Annotated[
    Optional[List[str]],
    typer.Option("--name", help="Names to filter by."),
]
FlagHostCurrent = Annotated[
    bool,
    typer.Option(
        "--current/--all",
        help="Show the current host.",
    ),
]
FlagHostNames = Annotated[
    Optional[List[str]],
    typer.Option("--name", help="Names to filter by."),
]
FlagTokenRequired = Annotated[str, typer.Option()]
FlagUUIDUser = Annotated[str, typer.Option()]


class ProfilesCommand(BaseTyperizable):

    typer_check_verbage = False
    typer_decorate = False
    typer_commands = dict(
        list="list",
        remove="remove",
        set="set",
        add="add",
        ls="list",
        rm="remove",
    )

    @classmethod
    def require_profile(
        cls,
        context: ContextData,
        config: Config | None = None,
    ) -> ProfileConfig:
        return (config if config is not None else context.config).profile

    @classmethod
    def handle_out(
        cls,
        context: ContextData,
        config: Config,
        *,
        show: bool = True,
        config_path_out: FlagConfigOut = None,
    ):
        if config_path_out is not None:
            context.console_handler.console.print("[green]Updating profile...")
            config.dump(config_path_out)
        elif not show:
            return
        else:
            context.console_handler.handle(
                handler_data=HandlerData(
                    data=config.model_dump_minimal(),
                    output_config=context.config.output,
                )
            )

    @classmethod
    def add(
        cls,
        _context: typer.Context,
        name: ArgumentProfileName,
        *,
        token: FlagTokenRequired,
        uuid: FlagUUIDUser,
        config_path: FlagConfig = None,
        config_path_out: FlagConfigOut = None,
    ):
        context = ContextData.resolve(_context)

        config = context.get_config(config_path)
        if name in config.profiles:
            context.console_handler.console.print(
                f"[red]Profile with name {name} already exists."
            )
            raise typer.Exit(0)

        config.profiles[name] = mwargs(ProfileConfig, uuid_user=uuid, token=token)
        cls.handle_out(context, config, config_path_out=config_path_out)

    @classmethod
    def set(
        cls,
        _context: typer.Context,
        name: ArgumentProfileName,
        *,
        token: flags.FlagTokenOptional = None,
        uuid: flags.FlagUUIDUserOptional = None,
        config_path: FlagConfig = None,
        config_path_out: FlagConfigOut = None,
    ):
        context = ContextData.resolve(_context)

        config = context.get_config(config_path)
        profile = cls.require_profile(context, config)

        if token is not None:
            profile.token = SecretStr(token)
        if uuid is not None:
            profile.uuid_user = uuid

        config.profiles[name] = profile
        cls.handle_out(context, config, show=False, config_path_out=config_path_out)

    @classmethod
    def remove(
        cls,
        _context: typer.Context,
        name: ArgumentProfileName,
        *,
        config_path: FlagConfig = None,
        config_path_out: FlagConfigOut = None,
    ):
        context = ContextData.resolve(_context)
        config = context.get_config(config_path)

        if name not in config.profiles:
            context.console_handler.console.print(f"[red]No such profile `{name}`.")
            raise typer.Exit(0)

        config.profiles.pop(name)
        cls.handle_out(context, config, config_path_out=config_path_out)

    @classmethod
    def list(
        cls,
        _context: typer.Context,
        *,
        current: FlagProfileCurrent = False,
        names: FlagProfileNames = None,
    ) -> None:
        """To set a ``profile``, see the ``use`` subcommand."""

        context = ContextData.resolve(_context)

        profiles = context.config.profiles
        data: Any
        if current:
            if names:
                CONSOLE.print("[red]Names cannot be specified when `--current` is.")
                raise typer.Exit(1)

            profile = cls.require_profile(context)
            data = (
                (
                    context.config.use.profile,
                    profile.model_dump(mode="json") if profile else profile,
                ),
            )
        else:
            data = ((pp, qq.model_dump(mode="json")) for pp, qq in profiles.items())
            if names is not None:
                data = ((pp, qq) for pp, qq in data if pp in names)

        context.console_handler.handle(
            handler_data=HandlerData(
                data=dict(data),
                output_config=context.config.output,
            )
        )

        return


class HostCommand(BaseTyperizable):
    typer_check_verbage = False
    typer_decorate = False
    typer_commands = dict(list="list", ls="list")

    @classmethod
    def list(
        cls,
        _context: typer.Context,
        current: FlagHostCurrent = False,
        names: FlagHostNames = None,
    ) -> None:
        """To set a ``host``, see the ``use`` subcommand."""
        context = ContextData.resolve(_context)

        hosts = context.config.hosts

        data: Any
        if current and names is None:
            host = context.config.host
            data = (
                (
                    context.config.use.host,
                    host.model_dump(mode="json") if host is not None else host,
                ),
            )
        else:
            data = ((pp, qq.model_dump(mode="json")) for pp, qq in hosts.items())
            if names is not None:
                data = ((pp, qq) for pp, qq in data if pp in names)

        context.console_handler.handle(
            handler_data=HandlerData(
                data=dict(data),
                output_config=context.config.output,
            )
        )
        return


class ConfigCommands(BaseTyperizable):
    typer_check_verbage = False
    typer_decorate = False
    typer_commands = dict(origin="origin", show="show", use="use")
    typer_children = dict(hosts=HostCommand, profiles=ProfilesCommand)

    @classmethod
    def use(
        cls,
        _context: typer.Context,
        host: Annotated[
            Optional[str],
            typer.Option("--host", "-h"),
        ] = None,
        profile: Annotated[
            Optional[str],
            typer.Option("--profile", "-p"),
        ] = None,
        config_path: FlagConfig = None,
        config_path_out: FlagConfigOut = None,
    ):
        context = ContextData.resolve(_context)
        config = context.get_config(config_path)

        if host is not None:
            if host not in config.hosts:
                msg = f"[red]No such host `{host}` in client config."
                CONSOLE.print(msg.format(host))
                raise typer.Exit(1)
            config.use.host = host

        if profile is not None:
            if profile not in config.profiles:
                msg = f"[red]No such profile `{profile}` in client config."
                CONSOLE.print(msg.format(host))
                raise typer.Exit(1)
            config.use.profile = profile

        if config_path_out is None:
            context.console_handler.handle(data=config.model_dump_minimal())
            raise typer.Exit(0)

        context.console_handler.console.print("[green]Updating client config.")
        config.dump(config_path_out)

    @classmethod
    def origin(cls, _context: typer.Context) -> None:
        context = ContextData.resolve(_context)
        yaml_datas = context.config.model_config["yaml_files"]
        assert isinstance(yaml_datas, dict)

        envvar = yaml_datas[util.PATH_CONFIG_CLIENT].get("envvar")
        assert envvar is not None

        print(environ.get(envvar, util.PATH_CONFIG_CLIENT))

    @classmethod
    def show(
        cls,
        _context: typer.Context,
        all_: Annotated[bool, typer.Option("--all/--minimal")] = False,
    ) -> None:
        context = ContextData.resolve(_context)
        config = context.config
        data = config.model_dump(mode="json") if all_ else config.model_dump_minimal()
        context.console_handler.handle(
            handler_data=HandlerData(
                data=data,
                output_config=context.config.output,
            )
        )


# NOTE: Should the docker stuff be optional? It could be added to the ``legere``
#       command as a plugin.
class DockerCommand(BaseTyperizable):
    typer_check_verbage = False
    typer_decorate = False
    typer_commands = dict(server="server", db="db")

    @classmethod
    def server(
        cls,
        _context: typer.Context,
        inspect: Annotated[
            bool,
            typer.Option(
                "-i",
                "--inspect",
                help="Display docker inspect results. Has priority over `--ips`.",
            ),
        ] = False,
        ips: Annotated[
            bool,
            typer.Option(
                "--ips",
                help="Display IPS from docker inspect results.",
            ),
        ] = False,
        config_path: FlagConfig = None,
        config_path_out: FlagConfigOut = None,
    ) -> None:
        import docker

        context = ContextData.resolve(_context)
        console = context.console_handler.console

        client = docker.DockerClient()  # type: ignore[attr-defined]
        if (container := client.containers.get("captura-server")) is None:
            console.print("[red]Docker compose project is not running.")
            raise typer.Exit(1)

        res = client.api.inspect_container(container.name)

        data: Dict[str, Any]
        match (inspect, ips):
            case (True, _):
                data = res
            case (_, True):
                networks = res["NetworkSettings"]["Networks"]
                data = {
                    network_name: network_detail["IPAddress"]
                    for network_name, network_detail in networks.items()
                }
            case _:
                networks = res["NetworkSettings"]["Networks"]
                hostconfs = {
                    "docker": HostConfig(
                        host="http://" + network_detail["IPAddress"] + ":8080",
                        remote=True,
                    )
                    for network_detail in networks.values()
                }
                if config_path:
                    with open(config_path, "r") as file:
                        config = Config.model_validate(yaml.safe_load(file))
                else:
                    config = context.config

                config.hosts.update(hostconfs)

                if config_path_out:
                    console.print("[green]Updating client config.")
                    with open(config_path_out, "w") as file:
                        data = config.model_dump_config()
                        yaml.dump(data, file)

                    raise typer.Exit()

                data = config.model_dump(
                    mode="json", exclude={"profile", "host", "token"}
                )
                context.console_handler.handle(
                    handler_data=HandlerData(
                        data=data,
                        output_config=context.config.output,
                    )
                )
                return

        context.console_handler.handle(
            handler_data=HandlerData(
                data=data,
                output_config=context.config.output,
            )
        )

    @classmethod
    def db(cls, _context: typer.Context):
        import docker

        context = ContextData.resolve(_context)
        console = context.console_handler.console

        client = docker.DockerClient()  # type: ignore
        if (container := client.containers.get("captura-db")) is None:
            console.print("[red]Docker compose project is not running.")
            raise typer.Exit(1)

        res = client.api.inspect_container(container.name)
        networks = res["NetworkSettings"]["Networks"]
        data = {
            network_name: network_detail["IPAddress"]
            for network_name, network_detail in networks.items()
        }
        context.console_handler.handle(
            handler_data=HandlerData(
                data=data,
                output_config=context.config.output,
            )
        )


__all__ = (
    "AssignmentRequests",
    "BaseRequests",
    "ContextData",
    "params",
    "CollectionRequests",
    "DocumentRequests",
    "GrantRequests",
    "TokenRequests",
    "UserRequests",
    "ConsoleHandler",
    "CONSOLE",
    "Requests",
    "Config",
)



# =========================================================================== #
from typing import Any, Dict

import httpx
import typer

# --------------------------------------------------------------------------- #
from captura.models import ChildrenCollection
from legere import flags
from legere.handlers import CONSOLE, AssertionHandler
from legere.requests.assignments import CollectionAssignmentRequests
from legere.requests.base import BaseRequests, ContextData, methodize, params


class CollectionRequests(BaseRequests):
    typer_commands = dict(
        read="req_read",
        create="req_create",
        delete="req_delete",
        update="req_update",
        search="req_search",
    )
    typer_children = dict(
        assignments=CollectionAssignmentRequests,
    )

    assignments: CollectionAssignmentRequests

    def __init__(
        self,
        context: ContextData,
        client: httpx.AsyncClient,
        *,
        handler: AssertionHandler | None = None,
        handler_methodize: bool = False,
    ):
        super().__init__(
            context,
            client,
            handler=handler,
            handler_methodize=handler_methodize,
        )
        self.assignments = CollectionAssignmentRequests.spawn_from(self)

    @property
    def a(self) -> CollectionAssignmentRequests:
        return self.assignments

    @classmethod
    def req_search(
        cls,
        _context: typer.Context,
        *,
        name_like: flags.FlagNameLike = None,
        description_like: flags.FlagDescriptionLike = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url("/collections"),
            params=params(
                name_like=name_like,
                description_like=description_like,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        uuid_collection: flags.ArgUUIDCollection,
        *,
        child: flags.FlagChildrenCollection | None = None,
        uuid_child: flags.FlagUUIDs = list(),
    ) -> httpx.Request:
        params: Dict[str, Any] = dict()
        match [child, not len(uuid_child)]:  # type: ignore[arg-type]
            case [None, True]:
                pass
            case [ChildrenCollection.documents, _]:
                params.update(uuid_document=uuid_child)
            case [ChildrenCollection.edits, _]:
                params.update(uuid_edit=uuid_child)
            case _:
                CONSOLE.print(
                    "[red]`--uuid-child` can only be used when `--child` is "
                    "provided."
                )
                raise typer.Exit(1)

        context = ContextData.resolve(_context)

        # Determine URL
        url_parts = ["collections", uuid_collection]
        if child is not None:
            url_parts.append(child)
        url = context.url("/" + "/".join(url_parts))
        return httpx.Request("GET", url, params=params, headers=context.headers)

    @classmethod
    def req_create(
        cls,
        _context: typer.Context,
        *,
        name: flags.FlagNameOptional = None,
        description: flags.FlagDescriptionOptional = None,
        public: flags.FlagPublicOptional = None,
        uuid_document: flags.FlagUUIDDocumentsOptional = list(),
        content: flags.FlagContentOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "POST",
            context.url("/collections"),
            params=dict(uuid_document=uuid_document),
            json=dict(
                name=name,
                description=description,
                public=public,
                content=content,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_delete(
        cls,
        _context: typer.Context,
        uuid_collection: flags.ArgUUIDCollection,
        *,
        content: flags.FlagContentOptional = None,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "DELETE",
            context.url(f"/collections/{uuid_collection}"),
            params=dict(force=force),
            headers=context.headers,
            content=content,
        )

    @classmethod
    def req_update(
        cls,
        _context: typer.Context,
        uuid_collection: flags.ArgUUIDCollection,
        *,
        name: flags.FlagNameOptional = None,
        description: flags.FlagDescriptionOptional = None,
        public: flags.FlagPublicOptional = None,
        content: flags.FlagContentOptional = None,
        uuid_user: flags.FlagUUIDUserOptional = None,
    ) -> httpx.Request:
        data = params(
            name=name,
            description=description,
            public=public,
            uuid_user=uuid_user,
            content=content,
        )
        context = ContextData.resolve(_context)
        return httpx.Request(
            "PATCH",
            context.url(f"/collections/{uuid_collection}"),
            json=data,
            headers=context.headers,
        )

    delete = methodize(req_delete, __func__=req_delete.__func__)  # type: ignore
    update = methodize(req_update, __func__=req_update.__func__)  # type: ignore
    create = methodize(req_create, __func__=req_create.__func__)  # type: ignore
    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore
    search = methodize(req_search, __func__=req_search.__func__)  # type: ignore


__all__ = ("CollectionRequests",)

if __name__ == "__main__":
    # --------------------------------------------------------------------------- #
    from legere.requests.base import typerize

    collections = typerize(CollectionRequests)
    collections()



# =========================================================================== #
import enum
from typing import Annotated, List, Optional

import httpx
import typer

# --------------------------------------------------------------------------- #
from legere.handlers import AssertionHandler

from .assignments import AssignmentRequests

# from .assignments import AssignmentRequests
from .base import BaseRequests, ContextData, params
from .collections import CollectionRequests
from .documents import DocumentRequests
from .events import EventsRequests
from .grants import GrantRequests
from .tokens import TokenRequests
from .users import UserRequests


# NOTE: All enums pertaining to tables should use the plural table names as
#       enum names.
class RequestsEnum(enum.Enum):
    users = UserRequests
    collections = CollectionRequests
    documents = DocumentRequests
    events = EventsRequests
    tokens = TokenRequests


class Requests(BaseRequests):
    typer_check_verbage = False
    typer_commands = dict(routes="req_routes", openapi="req_openapijson")
    typer_children = dict(
        assignments=AssignmentRequests,
        collections=CollectionRequests,
        documents=DocumentRequests,
        grants=GrantRequests,
        users=UserRequests,
        tokens=TokenRequests,
    )

    assignments: AssignmentRequests
    collections: CollectionRequests
    documents: DocumentRequests
    grants: GrantRequests
    users: UserRequests
    tokens: TokenRequests

    def __init__(
        self,
        context: ContextData,
        client: httpx.AsyncClient,
        *,
        handler: AssertionHandler | None = None,
        handler_methodize: bool = False,
    ):
        super().__init__(
            context,
            client,
            handler=handler,
            handler_methodize=handler_methodize,
        )
        self.assignments = AssignmentRequests.spawn_from(self)
        self.collections = CollectionRequests.spawn_from(self)
        self.documents = DocumentRequests.spawn_from(self)
        self.grants = GrantRequests.spawn_from(self)
        self.users = UserRequests.spawn_from(self)
        self.tokens = TokenRequests.spawn_from(self)

    @property
    def a(self) -> AssignmentRequests:
        return self.assignments

    @property
    def c(self) -> CollectionRequests:
        return self.collections

    @property
    def d(self) -> DocumentRequests:
        return self.documents

    @property
    def g(self) -> GrantRequests:
        return self.grants

    @property
    def u(self) -> UserRequests:
        return self.users

    @classmethod
    def req_routes(
        cls,
        _context: typer.Context,
        *,
        methods: Annotated[List[str] | None, typer.Option()] = None,
        names: Annotated[List[str] | None, typer.Option()] = None,
        names_fragment: Annotated[Optional[str], typer.Option] = None,
        paths_fragment: Annotated[Optional[str], typer.Option] = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url("/routes"),
            params=params(
                methods=methods,
                names=names,
                names_fragment=names_fragment,
                paths_fragment=paths_fragment,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_openapijson(
        cls,
        _context: typer.Context,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        res = context.req_openapijson()
        return res


__all__ = (
    "UserRequests",
    "CollectionRequests",
    "DocumentRequests",
    "GrantRequests",
    "AssignmentRequests",
    "EventsRequests",
    "RequestsEnum",
    "BaseRequests",
    "Requests",
)



# =========================================================================== #
from http import HTTPMethod
from typing import Any, Dict, Self, Set, Tuple, Type, overload

from pydantic import BaseModel, model_validator
from rich.table import Table
from sqlalchemy import Delete as sqaDelete
from sqlalchemy import Update, and_, delete, false, or_, select, true, update
from sqlalchemy.orm import Session

# --------------------------------------------------------------------------- #
from captura import util
from captura.auth import Token
from captura.controllers.access import Access, WithAccess, with_access
from captura.controllers.base import (
    Data,
    DataResolvedAssignment,
    DataResolvedGrant,
    ResolvedAssignmentCollection,
    ResolvedAssignmentDocument,
    ResolvedCollection,
    ResolvedDocument,
    ResolvedEvent,
    ResolvedGrantDocument,
    ResolvedGrantUser,
    ResolvedObjectEvents,
    ResolvedUser,
)
from captura.models import (
    Assignment,
    Collection,
    Document,
    Event,
    Grant,
    KindEvent,
    KindObject,
    User,
    uuids,
)
from captura.schemas import mwargs

logger = util.get_logger(__name__)


class AssocData(BaseModel):
    uuid_target_none: Set[str]
    uuid_assoc_none: Set[str]

    uuid_target_active: Set[str]
    uuid_target_deleted: Set[str]

    uuid_assoc_active: Set[str]
    uuid_assoc_deleted: Set[str]

    @model_validator(mode="after")
    def validate_no_intersections(self) -> Self:
        assert not len(self.uuid_target_active & self.uuid_target_deleted)
        assert not len(self.uuid_target_active & self.uuid_target_none)
        assert not len(self.uuid_target_deleted & self.uuid_target_none)

        assert not len(self.uuid_assoc_active & self.uuid_assoc_deleted)
        assert not len(self.uuid_assoc_active & self.uuid_assoc_none)
        assert not len(self.uuid_assoc_deleted & self.uuid_assoc_none)

        assert len(self.uuid_target_active) == len(self.uuid_assoc_active)
        assert len(self.uuid_target_deleted) == len(self.uuid_assoc_deleted)

        return self

    def render(self) -> Table:
        table = Table(show_header=False, show_lines=False, pad_edge=False)
        table.add_column()
        table.add_column()
        table.add_row(
            "uuid_target_none",
            str(self.uuid_target_none),
            style="red",
        )
        table.add_row(
            "uuid_target_active",
            str(self.uuid_target_active),
            style="blue",
        )
        table.add_row(
            "uuid_target_deleted",
            str(self.uuid_target_deleted),
            style="green",
        )
        table.add_row(
            "uuid_assoc_none",
            str(self.uuid_assoc_none),
            style="red",
        )
        table.add_row(
            "uuid_assoc_active",
            str(self.uuid_assoc_active),
            style="blue",
        )
        table.add_row(
            "uuid_assoc_deleted",
            str(self.uuid_assoc_deleted),
            style="green",
        )
        return table

    def q_del(self, model_assoc: Type, force: bool, *, rm_active: bool = True):
        uuid_assoc_rm = self.uuid_assoc_active.copy() if rm_active else set()

        q_del: Update | sqaDelete
        if force:
            uuid_assoc_rm |= self.uuid_assoc_deleted
            q_del = delete(model_assoc).where(model_assoc.uuid.in_(uuid_assoc_rm))
        else:
            q_del = (
                update(model_assoc)
                .where(model_assoc.uuid.in_(uuid_assoc_rm))
                .values(deleted=True)
            )

        return q_del, uuid_assoc_rm


class Delete(WithAccess):
    """Perform deletions."""

    @property
    def event_common(self) -> Dict[str, Any]:
        return dict(**super().event_common, kind=KindEvent.delete)

    def _event(self, item: Event, info: Set[Tuple[KindObject, None | str]]) -> int:
        # Pruning events should not create any more events besides one marking
        # that the pruning was performed.

        meth = (
            self.session.delete
            if self.force
            else lambda child: setattr(child, "deleted", True)
        )
        _ = tuple(
            (meth(child), info.add((child.kind_obj, child.uuid_obj)))  # type: ignore[func-returns-value]
            for child in item.flattened()
        )

        return len(_)

    def event_event(
        self,
        data: Data[ResolvedEvent] | Data[ResolvedObjectEvents],
        n: int,
        info: Set[Tuple[KindObject, str]],
    ) -> Event:
        token_user = data.token_user or self.token_user
        event = Event(
            **self.event_common,
            kind_obj=KindObject.user,
            uuid_obj=token_user.uuid,
            detail=f"User pruned `{n}` events.",
        )

        specs: Tuple[Tuple[KindObject, str | None], ...]
        match data.data:
            case ResolvedEvent(events=events):
                specs = tuple((ee.kind_obj, ee.uuid_obj) for ee in events)
            case ResolvedObjectEvents(kind_obj=kind_obj, uuid_obj=uuid_obj):
                specs = ((kind_obj, uuid_obj),)
            case _:
                raise ValueError(f"Invalid data of kind `{data.kind}`.")

        ff = "Events pruned by pruning object of kind `{0}` with uuid `{1}`."
        event.children = [
            Event(
                **self.event_common,
                kind_obj=kind_obj,
                uuid_obj=uuid_obj,
                detail=(detail := ff.format(kind_obj, uuid_obj)),
                # Every object should have a record.
                children=[
                    Event(
                        **self.event_common,
                        kind_obj=collatoral_kind_obj,
                        uuid_obj=collatoral_uuid_obj,
                        detail=detail,
                    )
                    for collatoral_kind_obj, collatoral_uuid_obj in info
                ],
            )
            for kind_obj, uuid_obj in specs
        ]
        return event

    @overload
    def event(
        self,
        data: Data[ResolvedEvent],
        # commit: bool = False,
    ) -> Data[ResolvedEvent]: ...

    @overload
    def event(
        self,
        data: Data[ResolvedObjectEvents],
        # commit: bool = False,
    ) -> Data[ResolvedObjectEvents]: ...

    def event(
        self,
        data: Data[ResolvedEvent] | Data[ResolvedObjectEvents],
        # commit: bool = False,
    ) -> Data[ResolvedEvent] | Data[ResolvedObjectEvents]:
        obj_info = set()  # type: ignore[var-annotated]
        n = sum(self._event(item, obj_info) for item in data.data.events)  # type: ignore[misc]
        data.event = self.event_event(data, n, obj_info)
        data.data.delete = True
        return data

    def object_events(
        self,
        data: Data[ResolvedObjectEvents],
    ) -> Data[ResolvedObjectEvents]:
        _ = self.event(data)
        return data

    a_object_events = with_access(Access.d_object_events)(object_events)

    # ----------------------------------------------------------------------- #
    # NOTE: Assocs crap

    def split_assocs(
        self,
        data: DataResolvedGrant | DataResolvedAssignment,
    ) -> Tuple[AssocData, Type]:
        model_assoc = data.data.get_model("assoc")
        model_target = data.data.get_model("target")
        q_assoc = (
            select(model_assoc)
            .join(model_target)
            .where(
                or_(
                    model_assoc.uuid.in_(data.data.uuid_assoc),
                    and_(
                        model_target.uuid.in_(data.data.uuid_target),
                        getattr(model_assoc, "id_" + data.data.kind_source.name)
                        == data.data.source.id,
                    ),
                )
            )
        )
        uuid_target_attr = "uuid_" + data.data.kind_target.name

        # NOTE: Simplify this.
        def one(bool_) -> Tuple[Set[str], Set[str]]:
            """Returns a tuple of a set of uuids of inactive assignments and
            uuids of active targets.
            """
            q = q_assoc.where(model_assoc.deleted == bool_())
            items = tuple(self.session.scalars(q))
            items = tuple(
                (item.uuid, getattr(item, uuid_target_attr)) for item in items
            )
            if not items:
                return (set(), set())
            return tuple(set(item) for item in zip(*items))  # type: ignore[return-value]

        deleted, active = (one(true), one(false))
        return (
            AssocData(
                uuid_assoc_none=data.data.uuid_assoc - deleted[0] - active[0],
                uuid_target_none=data.data.uuid_target - deleted[1] - active[1],
                uuid_assoc_deleted=deleted[0],
                uuid_target_deleted=deleted[1],
                uuid_assoc_active=active[0],
                uuid_target_active=active[1],
            ),
            model_assoc,
        )

    def assoc(
        self,
        data: DataResolvedAssignment | DataResolvedGrant,
    ) -> AssocData:
        assoc_data, model_assoc = self.split_assocs(data)
        q_del, uuid_assoc_rm = assoc_data.q_del(model_assoc, self.force)

        q_assocs = model_assoc.q_uuid(uuid_assoc_rm)
        assocs = tuple(self.session.scalars(q_assocs))

        self.session.execute(q_del)
        data.data.delete = self.force
        data.event = self.create_event_assoc(data, assocs)

        return assoc_data

    def create_event_assoc(
        self,
        data: DataResolvedGrant | DataResolvedAssignment,
        assocs: Tuple[Grant, ...] | Tuple[Assignment, ...],
    ) -> Event:
        # NOTE: Base event indicates the document, secondary event
        #       indicates the users for which permissions were revoked.
        #       Tertiary event indicates information about the association
        #       object. The shape of the tree is based off of the shape of
        #       the url on which this function can be called, where the
        #       document is first, the users are second, and the grants
        #       exist only as JSON.
        event_common = self.event_common
        kind_target = data.data.kind_target
        kind_assoc, kind_source = data.data.kind_assoc, data.data.kind_source

        uuid_target_attr_name = f"uuid_{kind_target.name}"
        detail = f"`{kind_assoc.name}`s deleted via `{kind_source.name}`."

        return Event(
            **event_common,
            uuid_obj=data.data.uuid_source,
            kind_obj=kind_source,
            detail=detail,
            children=[
                Event(
                    **event_common,
                    kind_obj=kind_target,
                    uuid_obj=getattr(assoc, uuid_target_attr_name),
                    detail=detail,
                    children=[
                        Event(
                            **event_common,
                            kind_obj=kind_assoc,
                            uuid_obj=assoc.uuid,
                            detail=detail,
                        )
                    ],
                )
                for assoc in assocs
            ],
        )

    # ----------------------------------------------------------------------- #

    def assignment_collection(
        self,
        data: Data[ResolvedAssignmentCollection],
    ) -> Data[ResolvedAssignmentCollection]:
        self.assoc(data)
        return data

    def assignment_document(
        self, data: Data[ResolvedAssignmentDocument]
    ) -> Data[ResolvedAssignmentDocument]:
        self.assoc(data)
        return data

    a_assignment_document = with_access(Access.d_assignment_document)(
        assignment_document
    )
    a_assignment_collection = with_access(Access.d_assignment_collection)(
        assignment_collection
    )

    # ----------------------------------------------------------------------- #
    # Grants

    def grant_user(
        self,
        data: Data[ResolvedGrantUser],
    ) -> Data[ResolvedGrantUser]:
        self.assoc(data)
        return data

    def grant_document(
        self, data: Data[ResolvedGrantDocument]
    ) -> Data[ResolvedGrantDocument]:
        self.assoc(data)
        return data

    a_grant_document = with_access(Access.d_grant_document)(grant_document)
    a_grant_user = with_access(Access.d_grant_user)(grant_user)

    # ----------------------------------------------------------------------- #

    def _user(
        self,
        data: Data[ResolvedUser],
        user: User,  # , commit: bool = False
    ) -> Data[ResolvedUser]:
        logger.debug("Deleting user `%s`.", user.uuid)
        session = self.session

        # Cleanup documents that only this user owns.
        documents_exclusive: Tuple[Document, ...] = tuple(
            session.execute(user.q_select_documents_exclusive()).scalars()
        )
        data_documents = mwargs(
            Data[ResolvedDocument],
            token_user=self.token_user,
            data=mwargs(
                ResolvedDocument,
                documents=documents_exclusive,
                token_user_grants=dict(),
            ),
        )
        self.document(data_documents)
        data.add(data_documents)

        # Cleanup collections.
        q_collections = user.q_collections(exclude_deleted=False)

        collections: Tuple[Collection, ...]
        collections = tuple(self.session.execute(q_collections).scalars())
        data_collections = mwargs(
            Data[ResolvedCollection],
            token_user=self.token_user,
            data=mwargs(
                ResolvedCollection,
                collections=collections,
            ),
        )
        self.collection(data_collections)
        data.add(data_collections)

        # NOTE: Should be done by `data.commit`.

        # if self.force:
        #     session.delete(User)
        # else:
        #     user.deleted = True
        #     session.add(user)

        # Event
        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.user,
            uuid_obj=user.uuid,
            children=[
                ee
                for data_item in (data_documents, data_collections)
                if (ee := data_item.event) is not None
            ],
        )
        # data.data.commit(session, commit, delete=self.force)

        return data

    def user(
        self,
        data: Data[ResolvedUser],  # commit: bool = False
    ) -> Data[ResolvedUser]:
        users = data.data.users

        data_users = tuple(self._user(data, user) for user in users)
        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.bulk,
            uuid_obj=None,
            children=[dd.event for dd in data_users],
        )
        data.add(*data_users)
        # data.data.commit(self.session, commit)

        return data

    a_user = with_access(Access.d_user)(user)

    # ----------------------------------------------------------------------- #

    def _collection(
        self,
        data: Data[ResolvedCollection],
        collection: Collection,
        # commit: bool = False,
    ) -> Data[ResolvedAssignmentCollection]:
        q = collection.q_select_documents(exclude_deleted=not self.force)
        documents = tuple(self.session.execute(q).scalars())
        q_assignments = (
            select(Assignment)
            .join(Document)
            .where(
                Document.uuid.in_(uuids(documents)),  # type: ignore[type-var]
                Assignment.id_collection == collection.id,
            )
        )
        assignments = {
            aa.uuid_document: aa for aa in self.session.scalars(q_assignments)
        }

        # Delete assigns and get events before deletion.
        data_assignments = mwargs(
            Data,
            token_user=data.token_user,
            data=mwargs(
                ResolvedAssignmentCollection,
                collection=collection,
                documents=documents,
                assignments=assignments,
            ),
        )
        if len(documents):
            _ = self.assignment_collection(data_assignments)

        data.add(data_assignments)

        # NOTE: Hard deletion is best done this way due to session.deleted
        #       strageness.
        if not self.force:
            collection.deleted = True
        else:
            data.data.delete = True
            self.session.delete(collection)

        # Create event
        event_assignments = data_assignments.event
        data_assignments.event = Event(
            **self.event_common,
            kind_obj=KindObject.assignment,
            uuid_obj=collection.uuid,
            detail="Bulk deletion of `assignment`s.",
        )
        if event_assignments is not None:
            data_assignments.event.children.append(event_assignments)

        # data.data.commit(self.session, commit)
        return data_assignments

    def collection(
        self,
        data: Data[ResolvedCollection],
        # commit: bool = False,
    ) -> Data[ResolvedCollection]:
        collections = data.data.collections
        if not (n := len(collections)):
            return data

        logger.debug("Deleting `%s` collections.", n)
        data_assignments = tuple(
            self._collection(data, collection) for collection in collections
        )
        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.collection,
            uuid_obj=None,
            children=[dd.event for dd in data_assignments],
            detail="Bulk deletion of `collection`s.",
        )

        # data.commit(self.session, commit)
        return data

    a_collection = with_access(Access.collection)(collection)

    # ----------------------------------------------------------------------- #

    def _document(
        self,
        data: Data[ResolvedDocument],
        document: Document,
        # *,
        # commit: bool = False,
    ) -> None:
        # Delete grants
        session = self.session
        if not self.force:
            document.deleted = True
        else:
            data.data.delete = True
            self.session.delete(document)

        q_users = document.q_select_users(
            exclude_deleted=not self.force, exclude_pending=False, pending=False
        )
        if users := tuple(session.execute(q_users).scalars()):
            data_grant = mwargs(
                Data[ResolvedGrantDocument],
                token_user=self.token_user,
                data=mwargs(
                    ResolvedGrantDocument,
                    token_user_grants=data.data.token_user_grants,
                    grants=data.data.token_user_grants,
                    document=document,
                    users=users,
                ),
            )
            self.grant_document(data_grant)
            data.add(data_grant)

        q_collections = document.q_select_collections(
            exclude_deleted=not self.force,
        )
        if collections := tuple(session.scalars(q_collections)):
            # NOTE: No deletion, etc. filtering as this is accoplished via
            #       ``q_select_collections``.
            uuid_collections = Collection.resolve_uuid(self.session, collections)
            q_assignments = (
                select(Assignment)
                .join(Collection)
                .where(
                    Assignment.id_document == document.id,
                    Collection.uuid.in_(uuid_collections),
                )
            )
            assignments = {
                aa.uuid_collection: aa for aa in session.scalars(q_assignments)
            }
            data_assignment = mwargs(
                Data[ResolvedAssignmentDocument],
                token_user=self.token_user,
                data=mwargs(
                    ResolvedAssignmentDocument,
                    document=document,
                    collections=collections,
                    assignments=assignments,
                ),
            )
            self.assignment_document(data_assignment)
            data.add(data_assignment)

    def document(
        self,
        data: Data[ResolvedDocument],
        # commit: bool = False,
    ) -> Data[ResolvedDocument]:
        documents = data.data.documents
        if not (n := len(documents)):
            return data

        logger.debug("Deleting `%s` collections", n)
        for document in data.data.documents:
            self._document(data, document)

        data.event = Event(
            **self.event_common,
            kind_obj=KindObject.bulk,
            children=[data_child.event for data_child in data.children],
        )
        data.data.delete = True
        return data

    a_document = with_access(Access.d_document)(document)


class WithDelete(WithAccess):
    delete: Delete

    def __init__(
        self,
        session: Session,
        token: Token | Dict[str, Any] | None,
        method: HTTPMethod | str,
        *,
        api_origin: str,
        force: bool = False,
        access: Access | None = None,
        delete: Delete | None = None,
    ):
        super().__init__(
            session,
            token,
            method,
            api_origin=api_origin,
            force=force,
            access=access,
        )

        if delete is None:
            delete = self.then(
                Delete,
                api_origin=api_origin,
                force=force,
                access=self.access,
            )
        self.delete = delete



# =========================================================================== #
import enum
import secrets
import uuid
from datetime import datetime
from typing import Annotated, Any, Callable, Dict, Set, Type, TypeAlias

from pydantic import BeforeValidator, Field

LENGTH_NAME: int = 96
LENGTH_TITLE: int = 128
LENGTH_DESCRIPTION: int = 256
LENGTH_URL: int = 256


class Level(enum.Enum):
    # NOTE: Must be consisten with sql, indexes start at 1
    view = 1
    modify = 2
    own = 3

    @classmethod
    def resolve(cls, v: "ResolvableLevel") -> "Level":
        match v:
            case LevelStr() as lvlstr:
                return Level[lvlstr.name]
            case Level() as self:
                return self
            case bad:
                raise ValueError(f"Cannot resolve level for `{bad}`.")


class LevelStr(str, enum.Enum):
    view = "view"
    modify = "modify"
    own = "own"


ResolvableLevel: TypeAlias = str | Level | LevelStr


class LevelHTTP(enum.Enum):
    DELETE = Level.own
    PUT = Level.modify
    POST = Level.modify
    PATCH = Level.modify
    GET = Level.view


class KindEvent(str, enum.Enum):
    create = "create"
    upsert = "upsert"
    update = "update"
    delete = "delete"
    grant = "grant"
    restore = "restore"


class PendingFrom(enum.Enum):
    created = 3
    granter = 2
    grantee = 1


class PendingFromStr(str, enum.Enum):
    created = "created"
    granter = "granter"
    grantee = "grantee"


# NOTE: This maps table names to their corresponding API names. It is important
#       to note that this uses singular names and not plural names.
class KindObject(str, enum.Enum):
    bulk = "bulk"
    user = "users"
    document = "documents"
    collection = "collections"
    edit = "edits"
    event = "events"
    assignment = "assignments"
    grant = "grants"


class Plural(str, enum.Enum):
    user = "users"
    document = "documents"
    collection = "collections"
    edit = "edits"
    event = "events"
    assignment = "assignments"
    grant = "grants"


class Singular(str, enum.Enum):
    events = "event"
    users = "user"
    documents = "document"
    collections = "collection"
    edits = "edit"
    assignments = "assignment"
    grants = "grant"


class KindRecurse(str, enum.Enum):
    depth_first = "depth-first"
    bredth_first = "bredth_first"


class ChildrenUser(str, enum.Enum):
    users = "users"
    collections = "collections"
    documents = "documents"
    edits = "edits"


class ChildrenCollection(str, enum.Enum):
    documents = "documents"
    edits = "edits"


class ChildrenDocument(str, enum.Enum):
    edits = "edits"


class ChildrenAssignment(str, enum.Enum):
    documents = "documents"
    collections = "collections"


class ChildrenGrant(str, enum.Enum):
    documents = "documents"
    users = "users"


FieldUUID = Annotated[
    str,
    _FieldUUID := Field(
        length=32,
        # min_length=32,
        # max_length=32,
        description="Universally unique identifier for an object.",
        examples=[uuid.uuid4()],
    ),
]
FieldUUIDOptional = Annotated[str | None, _FieldUUID]

FieldUnixTimestamp = Annotated[
    None | int,
    Field(
        description="Unix timestamp.",
        examples=[int(datetime.now().timestamp())],
    ),
]
FieldUnixTimestampOptional: TypeAlias = Annotated[
    datetime | None,
    Field(
        default=None,
        description="Optional unix timestamp.",
    ),
]

FieldName = Annotated[
    str,
    Field(
        min_length=1,
        max_length=LENGTH_NAME,
        description="Object name.",
        examples=["New Mexican Recipes", "Trails", "Software"],
    ),
]
FieldDescription = Annotated[
    str,
    Field(
        min_length=1,
        max_length=LENGTH_DESCRIPTION,
        description="Object description",
        examples=["This is a document/user/collection description."],
    ),
]
FieldUrl = Annotated[
    str | None,
    Field(
        min_length=8,
        max_length=LENGTH_URL,
        description="User url.",
        examples=["https://github.com/acederberg"],
    ),
]

EXAMPLE_CONTENT = (
    "orem ipsum dolor sit amet, consectetur adipiscing elit, sed do "
    "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut "
    "enim ad minim veniam, quis nostrud exercitation ullamco laboris "
    "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor "
    "in reprehenderit in voluptate velit esse cillum dolore eu fugiat "
    "nulla pariatur. Excepteur sint occaecat cupidatat non proident, "
    "sunt in culpa qui officia deserunt mollit anim id est laborum."
)

FieldUUIDS = Annotated[
    Set[str] | None,
    Field(
        default=None,
        description="UUIDs to filter with.",
        examples=[set(secrets.token_urlsafe() for _ in range(25))],
    ),
]
FieldID = Annotated[int, Field(description="Primary key(s) for tables.")]

FieldPending = Annotated[
    bool,
    Field(description="Grant awaiting approval (or not)."),
]
# FieldPendingFrom = Annotated[
#     PendingFrom,
#     Field(description="Grant initiator."),
# ]
FieldDeleted = Annotated[
    bool,
    Field(description="Object pending deletion or not."),
]
FieldDetail = Annotated[
    str | None,
    Field(default=None, description="Event detail."),
]
FieldLimitOptional: TypeAlias = Annotated[int | None, Field(default=None)]
FieldNameLike = Annotated[
    str | None,
    Field(
        description="Search objects for a name like this.",
        examples=["foobar", "billy mays"],
        default=None,
    ),
]
FieldDescriptionLike = Annotated[
    str | None,
    Field(
        description="Search objects for a description like this.",
        examples=["dolor sit amit,"],
        default=None,
    ),
]


def int_enum_from_name(
    EnumSubclass: Type[enum.Enum], callback: Callable[[Any], Any] | None = None
) -> Callable[[Any], Any]:
    def wrapper(v: Any) -> Any:
        match v:
            case EnumSubclass():  # type: ignore
                return v
            case str() as v_str:
                return EnumSubclass[v_str]
            case int() as v_int:
                return EnumSubclass(v_int)

        if callback is not None:
            v = callback(v)
        return v

    return wrapper


def int_enum_from_name_callback(v) -> None | Level:
    if isinstance(v, LevelStr):
        return FieldLevel(v.name)
    return v


FieldLevel = Annotated[
    Level,
    Field(description="Access level for grant."),
    BeforeValidator(int_enum_from_name(Level, int_enum_from_name_callback)),
]

FieldPendingFrom = Annotated[
    PendingFrom,
    Field(description="Grant pending origin."),
    BeforeValidator(int_enum_from_name(PendingFrom)),
]

FieldKindEvent = Annotated[
    KindEvent | None,
    Field(
        default=None,
        description="Event opperation tag.",
    ),
]

FieldKindObject = Annotated[
    KindObject | None,
    Field(
        default=None,
        description="Target object type. Often this is the name of corresponding table.",
        examples=[
            KindObject.user,
            KindObject.collection,
            KindObject.document,
        ],
    ),
]
FieldEmail = Annotated[
    str,
    Field(description="User email.", examples=["example1234@some.io"]),
]


FieldContent: TypeAlias = Annotated[
    Dict[str, Any] | None,
    Field(description="Document content.", default=None),
]


__all__ = (
    "ChildrenAssignment",
    "ChildrenCollection",
    "ChildrenDocument",
    "ChildrenGrant",
    "ChildrenUser",
    "KindEvent",
    "KindObject",
    "KindRecurse",
    "Level",
    "LevelHTTP",
    "LevelStr",
    "PendingFrom",
    "Plural",
    "Singular",
)



# =========================================================================== #
from typing import Annotated, Optional, Set, TypeAlias

from fastapi import Depends, HTTPException, Path, Query

# --------------------------------------------------------------------------- #
# from captura.fields import FieldLevel, KindEvent, KindObject, KindRecurse, LevelStr
from captura import fields

PathUUIDUser: TypeAlias = Annotated[str, Path(description="User uuids.")]
PathUUIDCollection: TypeAlias = Annotated[str, Path(description="Collection uuid.")]
PathUUIDDocument: TypeAlias = Annotated[str, Path(description="Document uuid.")]
PathUUIDEvent: TypeAlias = Annotated[str, Path(description="Event uuid.")]
PathUUIDObj: TypeAlias = Annotated[str, Path(description="Arbitrary object uuid.")]
PathKindObj: TypeAlias = Annotated[
    fields.FieldKindEvent, Path(description="Object kind.")
]

# --------------------------------------------------------------------------- #

# NOTE: Due to how `q_conds_` works (with empty `set()` versus `None`) the
#       empty set cannot be allowed. When nothing is passed it ought to be
#       `None`.
QueryUUIDCollectionOptional: TypeAlias = Annotated[
    Optional[Set[str]],
    Query(
        min_length=1,
        description="Optional collection uuids to filter by.",
    ),
]


def query_uuid_collection(
    uuid_collection: QueryUUIDCollectionOptional = None,
) -> Set[str]:
    # NOTE: See `uuid_user`.
    if uuid_collection is None:
        raise HTTPException(
            422, detail="Missing required `uuid_collection` query parameter."
        )
    return uuid_collection


QueryUUIDCollection: TypeAlias = Annotated[
    Optional[Set[str]], Depends(query_uuid_collection)
]

# --------------------------------------------------------------------------- #

QueryUUIDDocumentOptional: TypeAlias = Annotated[
    Set[str] | None,
    Query(
        min_length=1,
        description="Uuids to filter by.",
    ),
]


def query_uuid_document(uuid_document: QueryUUIDDocumentOptional = None) -> Set[str]:
    # NOTE: See `uuid_user`.
    if uuid_document is None:
        raise HTTPException(
            422, detail="Missing required `uuid_document` query parameter."
        )
    return uuid_document


QueryUUIDDocument: TypeAlias = Annotated[
    Optional[Set[str]], Depends(query_uuid_document)
]

# --------------------------------------------------------------------------- #

QueryUUIDUserOptional: TypeAlias = Annotated[
    Optional[Set[str]],
    Query(
        min_length=1,
        description="Optional user uuids to filter by.",
    ),
]


def query_uuid_user(uuid_user: QueryUUIDUserOptional = None) -> Set[str]:
    # NOTE: This is done because urls with no parameters will result in a 500.
    #       The message recieved is:

    #       .. code:: stdout
    #
    #           fastapi.exceptions.RequestValidationError: [
    #               {'type': 'set_type', 'loc': ('query', 'uuid_user'),
    #                'msg': 'Input should be a valid set',
    #                'input': PydanticUndefined,
    #                'url': 'https://errors.pydantic.dev/2.6/v/set_type'}
    #           ]
    if uuid_user is None:
        raise HTTPException(422, "Missing required `uuid_user` query parameter.")
    return uuid_user


QueryUUIDUser: TypeAlias = Annotated[Optional[Set[str]], Depends(query_uuid_user)]

# --------------------------------------------------------------------------- #

QueryForce: TypeAlias = Annotated[
    bool, Query(description="When true, objects cannot be restored.")
]
QueryLevel: TypeAlias = Annotated[fields.LevelStr, Query()]
QueryLevelOptional: TypeAlias = Annotated[fields.LevelStr | None, Query()]
QueryPendingFromOptional: TypeAlias = Annotated[fields.PendingFromStr | None, Query()]



# =========================================================================== #
import functools
from typing import Callable, Concatenate, List, Literal, ParamSpec, Type, TypeVar

from fastapi import Depends, HTTPException
from pydantic import TypeAdapter

# --------------------------------------------------------------------------- #
from captura.auth import TokenPermissionTier
from captura.controllers.access import Access, WithAccess
from captura.controllers.base import Data, ResolvedEvent, ResolvedObjectEvents
from captura.depends import DependsAccess, DependsDelete
from captura.models import (
    Event,
    KindEvent,
    KindObject,
)
from captura.schemas import (
    AsOutput,
    EventExtraSchema,
    EventMetadataSchema,
    EventParams,
    EventSchema,
    EventSearchSchema,
    OutputWithEvents,
    UserExtraSchema,
    mwargs,
)
from captura.views import args
from captura.views.base import (
    BaseView,
    OpenApiResponseCommon,
    OpenApiResponseUnauthorized,
    OpenApiTags,
)

# Decorators.
#
# Typing voodoo (from hell).

# NOTE: Replace this garbage with ``DependsAdminOnly``.
T_admin_only_controller = TypeVar(
    "T_admin_only_controller",
    bound=Access | WithAccess,
)
T_admin_only_return = TypeVar("T_admin_only_return")
T_admin_only_self = TypeVar("T_admin_only_self", bound=Type[BaseView])
P_admin_only = ParamSpec("P_admin_only")
CallableControllerFirst = Callable[
    Concatenate[T_admin_only_self, T_admin_only_controller, P_admin_only],
    T_admin_only_return,
]


def admin_only(
    fn: CallableControllerFirst[
        T_admin_only_self, T_admin_only_controller, P_admin_only, T_admin_only_return
    ],
) -> CallableControllerFirst[
    T_admin_only_self, T_admin_only_controller, P_admin_only, T_admin_only_return
]:
    @functools.wraps(fn)
    def wrapper(
        cls: Type[BaseView],
        access: T_admin_only_controller,
        *args: P_admin_only.args,
        **kwargs: P_admin_only.kwargs,
    ):
        if not access.token.tier == TokenPermissionTier.admin:
            detail = "This endpoint is for admins only."
            raise HTTPException(403, detail=detail)
        return fn(cls, access, *args, **kwargs)  # type: ignore

    return wrapper


# NOTE: Mounted directly on app. Easier to maintain here.
class EventSearchView(BaseView):
    view_routes = dict(
        get_event_objects=dict(
            url="/events/{uuid_event}/objects",
            name="Get Objects for Event",
            tags=[OpenApiTags.events],
        ),
        get_user_events=dict(
            url="/users/{uuid_user}/events",
            name="Get Events for User",
            tags=[OpenApiTags.users],
        ),
        # get_document_events=dict(
        #     url="/documents/{uuid_document}/events",
        #     name="Get Events for Document",
        #     tags=[OpenApiTags.documents],
        # ),
        # get_collection_events=dict(
        #     url="/collections/{uuid_collection}/events",
        #     name="Get Events for Collection",
        #     tags=[OpenApiTags.collections],
        # ),
        # get_grant_events=dict(
        #     url="/grants/documents/{uuid_document}/users/{uuid_user}/events",
        #     name="Get Events for Grant",
        #     tags=[OpenApiTags.grants],
        # ),
        # get_assignment_events=dict(
        #     url="/assignments/documents/{uuid_document}/collections/{uuid_collection}/events",
        #     name="Get Events for Assignment",
        #     tags=[OpenApiTags.assignments],
        # ),
    )

    view_router_args = dict(
        responses={
            **OpenApiResponseCommon,
            **OpenApiResponseUnauthorized,
        }
    )

    @classmethod
    @admin_only
    def get_event_objects(
        cls,
        access: DependsAccess,
        uuid_event: args.PathUUIDEvent,
        param: EventParams = Depends(),
    ) -> AsOutput[EventExtraSchema]:
        """Restore a deletion from an event."""

        data: Data[ResolvedEvent] = access.d_event(uuid_event)
        event = data.data.events[0]

        if param.root:
            event = event.find_root(access.session)

        return mwargs(
            AsOutput[EventExtraSchema],
            data=EventExtraSchema.model_validate(event),
        )

    @classmethod
    @admin_only
    def get_user_events(
        cls,
        access: DependsAccess,
        uuid_user: args.PathUUIDUser,
        param: EventParams = Depends(),
    ) -> OutputWithEvents[UserExtraSchema]:
        data: Data[ResolvedObjectEvents]
        data = access.d_object_events(uuid_user, KindObject.user, param)
        return mwargs(
            OutputWithEvents[UserExtraSchema],
            events=data.data.events,
            data=UserExtraSchema.model_validate(data.data.obj),
        )

    # @classmethod
    # @admin_only
    # def get_document_events(
    #     cls,
    #     access: DependsAccess,
    #     read: DependsRead,
    #     uuid_document: args.PathUUIDUser,
    #     param: EventParams = Depends(),
    # ) -> OutputWithEvents[DocumentExtraSchema]:
    #     item, events = read.object_events(uuid_document, KindObject.document, param)
    #     return mwargs(
    #         OutputWithEvents[DocumentExtraSchema],
    #         events=events,
    #         data=DocumentExtraSchema.model_validate(item),
    #     )
    #
    # @classmethod
    # @admin_only
    # def get_collection_events(
    #     cls,
    #     access: DependsAccess,
    #     read: DependsRead,
    #     uuid_collection: args.PathUUIDUser,
    #     param: EventParams = Depends(),
    # ) -> OutputWithEvents[CollectionExtraSchema]:
    #     item, events = read.object_events(uuid_collection, KindObject.collection, param)
    #     return mwargs(
    #         OutputWithEvents[CollectionExtraSchema],
    #         events=events,
    #         data=CollectionExtraSchema.model_validate(item),
    #     )
    #
    # @classmethod
    # @admin_only
    # def get_grant_events(
    #     cls,
    #     access: DependsAccess,
    #     read: DependsRead,
    #     uuid_document: args.PathUUIDDocument,
    #     uuid_user: args.PathUUIDUser,
    #     param: EventParams = Depends(),
    # ) -> OutputWithEvents[GrantExtraSchema]:
    #     session = access.session
    #     document = Document.resolve(session, uuid_document)
    #     user = User.resolve(session, uuid_user)
    #     grants = Grant.resolve_from_target(session, user, {document.uuid})
    #     if len(grants) != 1:
    #         raise HTTPException(500)
    #     (grant,) = grants
    #
    #     item, events = read.object_events(grant, KindObject.grant, param)
    #     return mwargs(
    #         OutputWithEvents[GrantExtraSchema],
    #         events=events,
    #         data=GrantExtraSchema.model_validate(item),
    #     )
    #
    # @classmethod
    # @admin_only
    # def get_assignment_events(
    #     cls,
    #     access: DependsAccess,
    #     read: DependsRead,
    #     uuid_document: args.PathUUIDCollection,
    #     uuid_collection: args.PathUUIDDocument,
    #     param: EventParams = Depends(),
    # ) -> OutputWithEvents[AssignmentExtraSchema]:
    #     session = access.session
    #     document = Document.resolve(session, uuid_document)
    #     collection = Collection.resolve(session, uuid_collection)
    #     grants = Assignment.resolve_from_target(session, collection, {document.uuid})
    #
    #     if len(grants) != 1:
    #         raise HTTPException(500)
    #     (grant,) = grants
    #
    #     item, events = read.object_events(grant, KindObject.grant, param)
    #     return mwargs(
    #         OutputWithEvents[AssignmentExtraSchema],
    #         events=events,
    #         data=AssignmentExtraSchema.model_validate(item),
    #     )


class EventView(BaseView):
    """
    ..
        NOTE: Events should never return results in flattened format from
              anywhere besides here.
    """

    view_routes = dict(
        get_event=dict(
            url="/{uuid_event}",
            name="Get Event",
        ),
        get_events=dict(
            url="",
            name="Search Events",
        ),
        delete_prune_event=dict(
            url="/{uuid_event}",
            name="Prune Event",
        ),
        delete_prune_object_events=dict(
            url="/{kind_obj}/{uuid_obj}",
            name="Prune Object Events",
        ),
        patch_undo_event=dict(
            url="/{uuid_event}/objects",
            name="Restore from Event",
        ),
    )
    view_router_args = dict(
        tags=[OpenApiTags.events],
        responses={
            **OpenApiResponseCommon,
            **OpenApiResponseUnauthorized,
        },
    )

    @classmethod
    @admin_only
    def delete_prune_event(
        cls,
        access: DependsAccess,
        delete: DependsDelete,
        uuid_event: args.PathUUIDEvent,
        param: EventParams = Depends(),
    ) -> OutputWithEvents[EventSchema]:
        """Proceed with caution. Admin only.

        This will result in pruning of information of other objects.
        """

        session = access.session
        event = Event.resolve(session, uuid_event)
        if param.root:
            event = event.find_root()

        event_serial = EventSchema.model_validate(event)
        data = delete.event(
            mwargs(
                Data[ResolvedEvent],
                data=mwargs(ResolvedEvent, events=(event,)),
            )
        )

        return mwargs(
            OutputWithEvents[EventSchema],
            data=event_serial,
            events=[EventSchema.model_validate(data.event)],
        )

    @classmethod
    @admin_only
    def delete_prune_object_events(
        cls,
        access: DependsAccess,
        delete: DependsDelete,
        kind_obj: Literal["users", "documents", "collections"],
        uuid_obj: args.PathUUIDObj,
        param: EventParams = Depends(),
    ) -> AsOutput[EventSchema]:
        "Proceed with caution. Admin only."

        data = delete.a_object_events(
            uuid_obj,
            KindObject(kind_obj),
            param,
            exclude_deleted=False,
        )

        return mwargs(
            AsOutput[EventSchema],
            data=EventSchema.model_validate(data.event),
        )

    @classmethod
    @admin_only
    def get_event(
        cls,
        access: DependsAccess,
        uuid_event: args.PathUUIDEvent,
        param: EventParams = Depends(),
    ) -> AsOutput[EventSchema]:
        "Get event by **uuid_event**. Admin only."

        session = access.session
        event = Event.resolve(session, uuid_event)
        if param.root:
            event = event.find_root(session)

        return mwargs(AsOutput[EventSchema], data=EventSchema.model_validate(event))

    @classmethod
    def get_events(
        cls,
        access: DependsAccess,
        param_search: EventSearchSchema = Depends(),
    ) -> AsOutput[List[EventMetadataSchema]]:
        # NOTE: Build the search.
        kwargs_search = param_search.model_dump(exclude={"before", "after"})
        if access.token.tier != TokenPermissionTier.admin:
            kwargs_search.update(uuid_user=access.token_user.uuid)

        session = access.session
        q = Event.q_select_search(
            **kwargs_search,
            before=(
                int(param_search.before.timestamp())
                if param_search.before is not None
                else None
            ),
            after=int(param_search.after.timestamp()),
        )
        # util.sql(session, q)
        events = session.execute(q).scalars()

        data = TypeAdapter(List[EventMetadataSchema]).validate_python(events)
        return mwargs(AsOutput, data=data)

    @classmethod
    def patch_undo_event(
        cls,
        delete: DependsDelete,
        uuid_event: args.PathUUIDUser,
    ) -> OutputWithEvents[EventSchema]:
        """Restore a deletion from an event.

        Return the events and updated objects.
        """

        event = (
            delete.access.event(uuid_event, return_data=False)
            .check_kind(KindEvent.delete)
            .check_not_undone()
        )
        delete.access.token_user.check_can_access_event(event)

        event = (
            delete.access.event(event.find_root(), return_data=False)
            .check_kind(KindEvent.delete)
            .check_not_undone()
        )
        delete.access.token_user.check_can_access_event(event)

        data = delete.event(
            mwargs(
                Data[ResolvedEvent],
                data=mwargs(ResolvedEvent, events=(event,)),
            )
        )

        return mwargs(
            OutputWithEvents[EventSchema],
            data=EventSchema.model_validate(event),
            event=data.event,
        )

    # TODO: Finish this later when time exists for it.
    # @classmethod
    # async def ws_events(
    #     cls,
    #     websocket: WebSocket,
    #     sessionmaker: DependsAsyncSessionMaker,
    #     token: DependsToken,
    #     param: EventSearchSchema = Depends(),
    #     flatten: args.QueryFlat = True,
    #     kind_recurse: args.QueryKindRecurse = KindRecurse.depth_first,
    #     wait: Annotated[int, args.Query(gt=1, lt=60)] = 1,
    # ) -> None:
    #     WS_MAX_LIFETIME: int = 36000
    #
    #     await websocket.accept()
    #     timestamp_connection_start = int(datetime.timestamp(datetime.now()))
    #     lifetime = 0
    #
    #     while (lifetime := lifetime + wait) < WS_MAX_LIFETIME:
    #         # events = await cls.get_events(
    #         #     sessionmaker,
    #         #     token,
    #         #     param=param,
    #         #     flatten=flatten,
    #         #     kind_recurse=kind_recurse,
    #         # )
    #         # await websocket.send_json(events)
    #         # await asyncio.sleep(wait)
    #         text = await websocket.receive_text()
    #         await websocket.send(f"text = `{text}`.")
    #
    #     ...



# =========================================================================== #
from typing import Annotated, Any, Dict, Optional

import typer
import uvicorn.config
import yaml
from rich.console import Console
from rich.syntax import Syntax
from sqlalchemy import Engine
from sqlalchemy.orm import sessionmaker

# --------------------------------------------------------------------------- #
from captura.config import Config
from captura.depends import session_maker
from captura.views.auth import AuthViewAuth0

from . import util

CONSOLE = Console()
FlagRun = Annotated[bool, typer.Option("--run", "--exit")]
FlagDummies = Annotated[bool, typer.Option("--dummies")]
FlagRecreateTables = Annotated[bool, typer.Option("--recreate-tables")]
FlagForCoverage = Annotated[
    bool,
    typer.Option(
        "--for-coverage",
        help=(
            "To run with coverage collection use this flag in the "
            "coverage subcommand, for instance "
            "`coverage run --include ./src/captura captura run --for-coverage`."
        ),
    ),
]
FlagReload = Annotated[
    Optional[bool],
    typer.Option(
        "--reload/--no-reload",
        help="Overwrite the reload strategy from ``config.app.is_dev``",
    ),
]


CONFIG = Config()  # type: ignore


class Cli:
    config: Config
    sessionmaker: sessionmaker
    engine: Engine

    def __init__(self, config: Config | None = None):
        self.config = config if config is not None else CONFIG
        self.engine = self.config.engine()
        self.sesionmaker = session_maker(self.engine)

    def show_config(self, envvars: bool = False):
        if envvars:
            text = yaml.dump(
                dict(
                    path_static=util.PATH_STATIC,
                    path_config_app=util.PATH_CONFIG_APP,
                    path_config_client=util.PATH_CONFIG_CLIENT,
                    path_config_test_app=util.PATH_CONFIG_TEST_APP,
                    path_config_test_client=util.PATH_CONFIG_TEST_CLIENT,
                    path_config_log=util.PATH_CONFIG_LOG,
                    path_hooks=util.PATH_HOOKS,
                    verbose=util.VERBOSE,
                    verbose_httpexceptions=util.VERBOSE_HTTPEXCEPTIONS,
                    session_secret=util.SESSION_SECRET,
                )
            )
        else:
            text = yaml.dump(self.config.model_dump(mode="json"), indent=2)
        text = "---\n\n" + text
        CONSOLE.print(Syntax(text, "yaml", theme="fruity"))

    def run(
        self,
        *,
        # for_coverage: FlagForCoverage = False,
        reload: FlagReload = None,
    ):
        """This function can be run by invoking this module (e.g. python -m
        app) or by using the command installed by ``pyproject.toml``, ``app``.

        To get the IP address of the Captura server do

        .. code:: shell

            # List the processes so that name of the container running this
            # code is known
            export CONTAINER_NAME=$( \
                docker compose --file=docker/docker-compose.yaml \
                ps --format '{{ .Name }}' \
                | grep server \
            )
            export FORMAT='{{
                .NetworkSettings.Networks.docker_documents.IPAddress }}'
            export CONTAINER_IP=$( \
                docker inspect \
                --format=$FORMAT \
                $CONTAINER_NAME \
            )
            echo $CONTAINER_IP


        and verify:


        .. code:: shell
            curl "http://$CONTAINER_IP:8080"

        """
        reload = reload if reload is not None else self.config.app.is_dev

        # if reload:
        #     util.CONSOLE_APP.print("[red]Reload is broken.")
        #     raise typer.Exit(1)

        self._run(app="captura:app", reload=reload)

    def _run(
        self,
        app,
        reload: bool,
    ):
        import uvicorn

        kwargs: Dict[str, Any] = dict(
            port=self.config.app.uvicorn_port,
            host=self.config.app.uvicorn_host,
        )
        # NOTE: Only specify ``reload_dirs`` in reload mode because warnings.
        if reload:
            kwargs.update(
                reload=reload,
                reload_dirs=util.PATH_APP,
            )

        uvicorn.run(app, **kwargs)

    def registration_code(self, email: str, validate: Optional[str] = None):
        code = AuthViewAuth0.create_code(self.config, email)
        if validate is None:
            return print(code)

        if code == validate:
            CONSOLE.print("[green]Registration code is valid!")
            return

        CONSOLE.print("[red]Registration code is not valid!")


def main() -> None:
    cli = Cli()

    tt = typer.Typer()
    tt.command("run")(cli.run)
    tt.command("config")(cli.show_config)
    tt.command("registration-code")(cli.registration_code)
    tt()


# NOTE: THIS APPEARS TO BE THE BEST SPOT FOR THIS! Please see
#
#       .. code:: txt
#
#         https://github.com/encode/uvicorn/issues/491
#
#       I know that loading the config like this is an antipattern. However
#       it does not matter since this module SHOULD ONLY EVER BE INVOKED BY
#       THE COMMAND LINE! This modules logger must be declared only after the
#       configuration as it will not be equiped with the correct handlers
#       otherwise.
LOGGING_CONFIG, _ = util.setup_logging(CONFIG.app.logging_configuration_path)
uvicorn.config.LOGGING_CONFIG.update(LOGGING_CONFIG)
logger = util.get_logger(__name__)


if __name__ == "__main__":
    main()



# =========================================================================== #

import httpx
import typer

# --------------------------------------------------------------------------- #
from legere import flags
from legere.requests.base import BaseRequests, ContextData, methodize, params


class CollectionAssignmentRequests(BaseRequests):
    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        uuid_collection: flags.ArgUUIDCollection,
        *,
        uuid_document: flags.FlagUUIDDocumentsOptional = list(),
        limit: flags.FlagLimitOptional = None,
        randomize: bool = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url(f"/assignments/collections/{uuid_collection}"),
            params=params(
                uuid_document=uuid_document,
                limit=limit,
                randomize=randomize,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_delete(
        cls,
        _context: typer.Context,
        uuid_collection: flags.ArgUUIDCollection,
        *,
        uuid_document: flags.FlagUUIDDocuments,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "DELETE",
            context.url(f"/assignments/collections/{uuid_collection}"),
            params=dict(uuid_document=uuid_document, force=force),
            headers=context.headers,
        )

    @classmethod
    def req_create(
        cls,
        _context: typer.Context,
        uuid_collection: flags.ArgUUIDCollection,
        *,
        uuid_document: flags.FlagUUIDDocuments,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "POST",
            context.url(f"/assignments/collections/{uuid_collection}"),
            params=dict(uuid_document=uuid_document, force=force),
            headers=context.headers,
        )

    typer_commands = dict(
        read="req_read",
        create="req_create",
        delete="req_delete",
    )

    create = methodize(req_create, __func__=req_create.__func__)  # type: ignore
    delete = methodize(req_delete, __func__=req_delete.__func__)  # type: ignore
    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore


class DocumentAssignmentRequests(BaseRequests):
    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        uuid_collection: flags.FlagUUIDCollectionsOptional = list(),
        limit: flags.FlagLimitOptional = None,
        randomize: bool = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url(f"/assignments/documents/{uuid_document}"),
            params=params(
                uuid_collection=uuid_collection,
                limit=limit,
                randomize=randomize,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_delete(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        uuid_collection: flags.FlagUUIDCollections,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "DELETE",
            context.url(f"/assignments/documents/{uuid_document}"),
            params=dict(uuid_collection=uuid_collection, force=force),
            headers=context.headers,
        )

    @classmethod
    def req_create(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        uuid_collection: flags.FlagUUIDCollections,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "POST",
            context.url(f"/assignments/documents/{uuid_document}"),
            params=dict(uuid_collection=uuid_collection, force=force),
            headers=context.headers,
        )

    typer_commands = dict(
        read="req_read",
        create="req_create",
        delete="req_delete",
    )

    create = methodize(req_create, __func__=req_create.__func__)  # type: ignore
    delete = methodize(req_delete, __func__=req_delete.__func__)  # type: ignore
    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore


class AssignmentRequests(BaseRequests):
    typer_children = dict(
        collections=CollectionAssignmentRequests,
        documents=DocumentAssignmentRequests,
    )

    collections: CollectionAssignmentRequests
    documents: DocumentAssignmentRequests

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.documents = DocumentAssignmentRequests.spawn_from(self)
        self.collections = CollectionAssignmentRequests.spawn_from(self)

    @property
    def c(self) -> CollectionAssignmentRequests:
        return self.collections

    @property
    def d(self) -> DocumentAssignmentRequests:
        return self.documents


__all__ = (
    "CollectionAssignmentRequests",
    "DocumentAssignmentRequests",
    "AssignmentRequests",
)


if __name__ == "__main__":
    # --------------------------------------------------------------------------- #
    from legere.requests.base import typerize

    assignments = typerize(AssignmentRequests)
    assignments()



# =========================================================================== #
from typing import List

from pydantic import TypeAdapter
from sqlalchemy import func, select
from sqlalchemy.sql.expression import false

# --------------------------------------------------------------------------- #
from captura.depends import DependsAccess, DependsCreate, DependsDelete
from captura.models import Assignment, Collection, Document, Level
from captura.schemas import (
    AsOutput,
    AssignmentCreateSchema,
    AssignmentSchema,
    EventSchema,
    OutputWithEvents,
    mwargs,
)
from captura.views import args
from captura.views.base import BaseView, OpenApiResponseUnauthorized, OpenApiTags

OpenApiResponseAssignment = {
    **OpenApiResponseUnauthorized,
    403: dict(
        model=dict,
        description=(
            "Raised when a user does not own any of the collections in "
            "question or when a user cannot at least view a document in "
            "question."
        ),
    ),
}


# NOTE: Should mirron :class:`GrantView`. Updates not supported, scoped by
#       collection.
class DocumentAssignmentView(BaseView):
    adapter = TypeAdapter(List[AssignmentSchema])

    view_router_args = dict(
        tags=[OpenApiTags.assignments], responses=OpenApiResponseAssignment
    )

    view_routes = dict(
        delete_assignments_document=dict(
            url="/{uuid_document}",
            name="Remove Document from Collections by Deleting Assignments",
        ),
        post_assignments_document=dict(
            url="/{uuid_document}",
            name="Add Document to Collections by Creating Assignments",
        ),
        get_assignments_document=dict(
            url="/{uuid_document}",
            name="Read Assignments of Document to Collections",
        ),
    )

    @classmethod
    def delete_assignments_document(
        cls,
        delete: DependsDelete,
        uuid_document: args.PathUUIDDocument,
        uuid_collection: args.QueryUUIDCollection,
    ) -> OutputWithEvents[List[AssignmentSchema]]:
        """Remove document from some collections.

        Collection owners should remove this document from their collection
        using ``DELETE /assignments/collections/{uuid_collection}``.

        All that is required is a token and document ownership.

        Collection ownership is not enforced here so that the document owner
        have the final say in which collections their document can be included.

        Document ownership is enforced since only document owners can
        reject/accept their document from being part of a particular collection.
        """
        data = delete.access.d_assignment_document(
            uuid_document,
            uuid_collection,  # type: ignore
            allow_public=False,
            validate_collections=False,
            level=Level.own,
            exclude_deleted=not delete.force,
        )

        assignments, events = list(), list()
        if len(data.data.collections):
            delete.assignment_document(data)
            assignments = cls.adapter.validate_python(data.data.assignments.values())
            data.commit(delete.session)
            events.append(EventSchema.model_validate(data.event))

        return mwargs(
            OutputWithEvents[List[AssignmentSchema]],
            events=events,
            data=assignments,
        )

    @classmethod
    def post_assignments_document(
        cls,
        create: DependsCreate,
        uuid_document: args.PathUUIDDocument,
        uuid_collection: args.QueryUUIDCollection,
    ) -> OutputWithEvents[List[AssignmentSchema]]:
        """Add this document to collections.

        All that is required is a token and document ownership.

        Collection ownership is not enforced here so that the document owner
        have the final say in which collections their document can be included.

        Document ownership is enforced since only document owners can
        reject/accept their document from being part of a particular collection.
        """
        data = create.access.d_assignment_document(
            uuid_document,
            uuid_collection,  # type: ignore
            allow_public=False,
            level=Level.own,
            validate_collections=False,
        )
        create.create_data = AssignmentCreateSchema()
        data_final = create.assignment_document(data)
        data_final.commit(create.session)

        assignments = data_final.data.assignments
        return mwargs(
            OutputWithEvents[List[AssignmentSchema]],
            data=cls.adapter.validate_python(assignments.values()),
            events=[EventSchema.model_validate(data_final.event)],
        )

    @classmethod
    def get_assignments_document(
        cls,
        access: DependsAccess,
        uuid_document: args.PathUUIDDocument,
        *,
        uuid_collection: args.QueryUUIDCollectionOptional = None,
        limit: int | None = None,
        randomize: bool = False,
    ) -> AsOutput[List[AssignmentSchema]]:
        """Read the collections to which this document belongs.

        For non-private documents all that is required is a token.
        For private documents a grant of level view is required.
        """

        # NOTE: Check document access first. Then data is constructed later.
        #       This is done because get the collections can be expensive due
        #       to `ORDER BY RANDOM()`.
        document = access.document(
            uuid_document,
            allow_public=True,
            level=Level.view,
            return_data=False,
        )

        q = (
            select(Collection)
            .join(Assignment)
            .where(
                Assignment.id_document == document.id,
                Assignment.deleted == false(),
                Collection.deleted == false(),
            )
        )
        if uuid_collection is not None:
            q = q.where(Collection.uuid.in_(uuid_collection))
        if randomize:
            q = q.order_by(func.random())
        if limit:
            q = q.limit(limit)

        collections = tuple(access.session.scalars(q))
        data = access.d_assignment_document(
            document,
            collections,
            level=Level.view,
            validate_document=False,  # NOTE: Already validated.
            validate_collections=False,
            all_collections=False,
            allow_public=True,
        )

        # NOTE: So that the radomness of the collections is passed down here.
        assignments = tuple(
            ass
            for collection in collections
            if (ass := data.data.assignments.get(collection.uuid)) is not None
        )
        return mwargs(
            AsOutput[List[AssignmentSchema]],
            data=cls.adapter.validate_python(assignments),
        )


class CollectionAssignmentView(BaseView):
    adapter = TypeAdapter(List[AssignmentSchema])

    view_routes = dict(
        delete_assignments_collection=dict(
            url="/{uuid_collection}",
            description="Remove Documents from Collection (by Deleting Assignments)",
        ),
        post_assignments_collection=dict(
            url="/{uuid_collection}",
            description="Add Documents to Collection (by Creating Assignments)",
        ),
        get_assignments_collection=dict(
            url="/{uuid_collection}",
            description="Read Documents for Collection.",
        ),
    )
    view_router_args = dict(
        tags=[OpenApiTags.assignments], responses=OpenApiResponseAssignment
    )

    @classmethod
    def delete_assignments_collection(
        cls,
        delete: DependsDelete,
        uuid_collection: args.PathUUIDCollection,
        uuid_document: args.QueryUUIDDocument,
    ) -> OutputWithEvents[List[AssignmentSchema]]:
        data = delete.access.d_assignment_collection(
            uuid_collection,
            uuid_document,  # type: ignore
            allow_public=False,
            level=Level.own,
            validate_documents=False,
        )

        assignments, events = list(), list()
        if len(data.data.documents):
            delete.assignment_collection(data)
            assignments = cls.adapter.validate_python(data.data.assignments.values())
            data.commit(delete.session)
            events.append(EventSchema.model_validate(data.event))

        return mwargs(
            OutputWithEvents[List[AssignmentSchema]],
            events=events,
            data=assignments,
        )

    @classmethod
    def post_assignments_collection(
        cls,
        create: DependsCreate,
        uuid_collection: args.PathUUIDCollection,
        uuid_document: args.QueryUUIDDocument,
    ) -> OutputWithEvents[List[AssignmentSchema]]:
        data = create.access.d_assignment_collection(
            uuid_collection,
            uuid_document,  # type: ignore
            level=Level.view,
            allow_public=False,
            validate_documents=False,
        )
        create.create_data = AssignmentCreateSchema()
        data_final = create.assignment_collection(data)
        data_final.commit(create.session)

        return mwargs(
            OutputWithEvents[List[AssignmentSchema]],
            data=cls.adapter.validate_python(data_final.data.assignments.values()),
            events=[EventSchema.model_validate(data_final.event)],
        )

    @classmethod
    def get_assignments_collection(
        cls,
        access: DependsAccess,
        uuid_collection: args.PathUUIDCollection,
        *,
        uuid_document: args.QueryUUIDDocumentOptional = None,
        limit: int | None = None,
        randomize: bool = False,
    ) -> AsOutput[List[AssignmentSchema]]:
        collection = access.collection(uuid_collection, return_data=False)
        q = (
            select(Document)
            .join(Assignment)
            .where(
                Document.deleted == false(),
                Assignment.deleted == false(),
                Assignment.id_collection == collection.id,
            )
        )
        if uuid_document is not None:
            q = q.where(Document.uuid.in_(uuid_document))
        if randomize:
            q = q.order_by(func.random())
        if limit:
            q = q.limit(limit)

        documents = tuple(access.session.scalars(q))
        data = access.d_assignment_collection(
            collection,
            documents,
            validate_documents=False,
        )
        assignments = tuple(
            ass
            for document in documents
            if (ass := data.data.assignments.get(document.uuid)) is not None
        )
        return mwargs(
            AsOutput[List[AssignmentSchema]],
            data=cls.adapter.validate_python(assignments),
        )



__version__ = "0.1.6-beta"

from fastapi import FastAPI

from . import hooks, util
from .auth import Auth
from .config import Config
from .models import (
    AssocCollectionDocument,
    AssocUserDocument,
    ChildrenCollection,
    ChildrenDocument,
    ChildrenUser,
    Collection,
    Document,
    KindEvent,
    KindObject,
    Level,
    LevelStr,
    Tables,
    User,
)
from .views import AppView

app: FastAPI = AppView.view_router  # type: ignore

if util.PATH_HOOKS_USE:
    hooks.do_hooks(AppView)


__all__ = (
    # Application
    "Auth",
    "Config",
    "AppView",
    # Models
    "User",
    "Collection",
    "AssocCollectionDocument",
    "AssocUserDocument",
    "Document",
    "Tables",
    "KindEvent",
    "KindObject",
    "Level",
    "LevelStr",
    "KindObject",
    "ChildrenUser",
    "ChildrenCollection",
    "ChildrenDocument",
)



# =========================================================================== #

# =========================================================================== #

import httpx
import typer

# --------------------------------------------------------------------------- #
from legere import flags
from legere.handlers import AssertionHandler
from legere.requests.base import BaseRequests, ContextData, methodize, params, typerize
from legere.requests.grants import UserGrantRequests


class UserRequests(BaseRequests):
    @classmethod
    def req_search(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        child: flags.FlagChildrenUser = None,
        uuids: flags.FlagUUIDs = None,
        name_like: flags.FlagNameLike = None,
        description_like: flags.FlagDescriptionLike = None,
        limit: flags.FlagLimit = 10,
        include_public: flags.FlagIncludePublic = True,
        randomize: bool = True,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)

        #
        # if uuid_user is None:
        #     uuid_user = context.config.profile.uuid_user  # type: ignore

        child_name = child.name if child is not None else "users"
        return httpx.Request(
            "GET",
            context.url(f"/users/{uuid_user}/{child_name}"),
            headers=context.headers,
            params=params(
                uuids=uuids,
                limit=limit,
                name_like=name_like,
                description_like=description_like,
                include_public=include_public,
                randomize=randomize,
            ),
        )

    @classmethod
    def req_read(
        cls, _context: typer.Context, uuid_user: flags.ArgUUIDUser
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url(f"/users/{uuid_user}"),
            headers=context.headers,
        )

    @classmethod
    def req_update(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        name: flags.FlagNameOptional = None,
        description: flags.FlagDescriptionOptional = None,
        content: flags.FlagContentOptional = None,
        url: flags.FlagUrlOptional = None,
        url_image: flags.FlagUrlImageOptional = None,
        public: flags.FlagPublicOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        json = dict(
            name=name,
            description=description,
            url=url,
            url_image=url_image,
            public=public,
            content=content,
        )
        return httpx.Request(
            "PATCH",
            context.url(f"/users/{uuid_user}"),
            json=json,
            headers=context.headers,
        )

    @classmethod
    def req_delete(
        cls,
        _context: typer.Context,
        uuid_user: flags.ArgUUIDUser,
        *,
        force: flags.FlagForce = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "DELETE",
            context.url(f"/users/{uuid_user}"),
            params=dict(
                uuid_user=uuid_user,
                force=force,
            ),
            headers=context.headers,
        )

    def __init__(
        self,
        context: ContextData,
        client: httpx.AsyncClient,
        *,
        handler: AssertionHandler | None = None,
        handler_methodize: bool = False,
    ):
        super().__init__(
            context,
            client,
            handler=handler,
            handler_methodize=handler_methodize,
        )
        self.grants = UserGrantRequests.spawn_from(self)

    @property
    def g(self) -> UserGrantRequests:
        return self.grants

    typer_commands = dict(
        read="req_read",
        search="req_search",
        update="req_update",
        delete="req_delete",
    )
    typer_children = {"grants": UserGrantRequests}

    grants: UserGrantRequests
    update = methodize(req_update, __func__=req_update.__func__)  # type: ignore
    delete = methodize(req_delete, __func__=req_delete.__func__)  # type: ignore
    search = methodize(req_search, __func__=req_search.__func__)  # type: ignore
    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore


__all__ = ("UserRequests",)


if __name__ == "__main__":

    users = typerize(UserRequests)
    users()
